# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""pretty_element.py

Pretty-printing xml.etree.ElementTree.Element subclass
"""
from __future__ import absolute_import
import importlib
import re
import sys

from jss import tools


_DUNDER_PATTERN = re.compile(r'__[a-zA-Z]+__')
_RESERVED_METHODS = ('cached',)

# ElementTree monkey patch borrowed with love from Matteo Ferla.
# https://blog.matteoferla.com/2019/02/uniprot-xml-and-python-elementtree.html
sys.modules.pop('xml.etree.ElementTree', None)
sys.modules['_elementtree'] = None
ElementTree = importlib.import_module('xml.etree.ElementTree')


class PrettyElement(ElementTree.Element):
    """Pretty printing element subclass

    Element subclasses xml.etree.ElementTree.Element to pretty print.
    """

    def __init__(self, tag, attrib={}, **extra):
        if isinstance(tag, ElementTree.Element):
            super(PrettyElement, self).__init__(tag.tag, tag.attrib, **extra)
            self.text = tag.text
            self.tail = tag.tail
            self._children = [PrettyElement(child) for child in tag]
        else:
            super(PrettyElement, self).__init__(tag, attrib, **extra)

    # Replace standard Element.__str__ with our cache-aware
    # pretty-printing one.
    __str__ = tools.element_str

    def __getattr__(self, name):
        # Any dunder methods should be passed as is to the superclass.
        # There are also some method names which need to be assumed to
        # be from the superclass, lest we endlessly loop.
        if re.match(_DUNDER_PATTERN, name) or name in _RESERVED_METHODS:
            return super(PrettyElement, self).__getattr__(name)
        result = super(PrettyElement, self).find(name)
        if result is not None:
            return result
        else:
            raise AttributeError(
                'There is no element with the tag "{}"'.format(name))

    # TODO: This can be removed once `JSSObject.__init__` signature is fixed.
    def makeelement(self, tag, attrib):
        """Return a PrettyElement with tag and attrib."""
        # We have to override Element's makeelement, which uses the
        # class' __init__. Since python-jss objects override this and
        # repurpose it, instantiating a sub element with
        # ElementTree.SubElement or copy will fail.

        # This situation will be resolved when JSSObject stops
        # subclassing Element.
        return PrettyElement(tag, attrib)

    def append(self, item):
        super(PrettyElement, self).append(self._convert(item))

    def insert(self, index, item):
        super(PrettyElement, self).insert(index, self._convert(item))

    def extend(self, items):
        super(PrettyElement, self).extend(self._convert(item) for item in items)

    def _convert(self, item):
        """If item is not a PrettyElement, make it one"""
        return item if isinstance(item, PrettyElement) else PrettyElement(item)



#!/usr/bin/env python
# Copyright (C) 2014-2018 Shea G Craig, 2018 Mosen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""distribution_point.py

Classes representing the various types of file storage available to
the JAMF Pro Server.
"""

from __future__ import division
from __future__ import print_function

from __future__ import absolute_import
import os
import re
import shutil
import socket
import subprocess
import sys
import io
import math
import multiprocessing
import threading

sys.path.insert(0, '/Library/AutoPkg/JSSImporter')
import requests

try:
    # Python 2.6-2.7
    from HTMLParser import HTMLParser
except ImportError:
    # Python 3
    from html.parser import HTMLParser


# 2 and 3 compatible
try:
    from urllib.parse import urlparse, urlencode, unquote, quote
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode, unquote, quote
    from urllib2 import urlopen, Request, HTTPError

from . import casper
from . import abstract
from .exceptions import JSSError
try:
    from .contrib.mount_shares_better import mount_share
except ImportError:
    # mount_shares_better uses PyObjC. If using non-system python,
    # chances are good user has not set up PyObjC, so fall back to
    # subprocess to mount. (See mount methods).
    mount_share = None
from .tools import (is_osx, is_linux, is_package)

try:
    import boto.s3
    from boto.s3.connection import S3Connection, OrdinaryCallingFormat, S3ResponseError
    from boto.s3.key import Key

    BOTO_AVAILABLE = True
except ImportError:
    print("boto is not available, you will not be able to use the AWS distribution point type")
    BOTO_AVAILABLE = False

PKG_FILE_TYPE = '0'
EBOOK_FILE_TYPE = '1'
IN_HOUSE_APP_FILE_TYPE = '2'


def auto_mounter(original):
    """Decorator for automatically mounting, if needed."""
    def mounter(*args):
        """If not mounted, mount."""
        self = args[0]
        if not self.is_mounted():
            self.mount()
        return original(*args)
    return mounter


# pylint: disable=too-few-public-methods
class Repository(object):
    """Base class for file repositories.

    This class is not usable on its own; however, it provides the base
    init which all subclasses should use.

    Attributes:
        connection (dict): Dictionary for storing connection arguments.
        required_attrs (Set): A set of the keys which must be supplied to the initializer, otherwise a JSSError will
            be raised.

    Raises:
        JSSError: If mandatory arguments are not supplied to the initializer.
    """
    required_attrs = set()

    def __init__(self, **connection_args):
        """Store the connection information."""
        if self.required_attrs.issubset(set(connection_args.keys())):
            self.connection = connection_args
            self._build_url()
        else:
            missing_attrs = self.required_attrs.difference(
                set(connection_args.keys()))
            raise JSSError(
                "Missing REQUIRED argument(s) %s to %s distribution point." %
                (list(missing_attrs), self.__class__))

    def __repr__(self):
        """Return string representation of connection arguments."""
        output = ["Distribution Point: %s" % self.connection["url"]]
        output.append("Type: %s" % type(self))
        output.append("Connection Information:")
        for key, val in self.connection.items():
            output.append("\t%s: %s" % (key, val))

        return "\n".join(output) + "\n"

    def _build_url(self):
        """Private build url method."""
        raise NotImplementedError


# pylint: enable=too-few-public-methods

class FileRepository(Repository):
    """Local file shares."""

    def _build_url(self):
        """Build a connection URL."""
        pass

    def copy_pkg(self, filename, _):
        """Copy a package to the repo's Package subdirectory.

        Args:
            filename: Path for file to copy.
            _: Ignored. Used for compatibility with JDS repos.
        """
        basename = os.path.basename(filename)
        self._copy(filename, os.path.join(self.connection["mount_point"],
                                          "Packages", basename))

    def _copy(self, filename, destination):   # pylint: disable=no-self-use
        """Copy a file or folder to the repository.

        Will mount if needed.

        Args:
            filename: Path to copy.
            destination: Remote path to copy file to.
        """
        full_filename = os.path.abspath(os.path.expanduser(filename))

        if os.path.isdir(full_filename):
            shutil.copytree(full_filename, destination)
        elif os.path.isfile(full_filename):
            shutil.copyfile(full_filename, destination)

    def delete(self, filename):
        """Delete a file from the repository.

        Args:
            filename: String filename only (i.e. no path) of file to
                delete.
        """
        folder = "Packages"
        path = os.path.join(self.connection["mount_point"], folder, filename)
        if os.path.isdir(path):
            shutil.rmtree(path)
        elif os.path.isfile(path):
            os.remove(path)

    def exists(self, filename):
        """Report whether a file exists on the distribution point.

        Determines file type by extension.

        Args:
            filename: Filename you wish to check. (No path! e.g.:
                "AdobeFlashPlayer-14.0.0.176.pkg")
        """
        filepath = os.path.join(
            self.connection["mount_point"], "Packages", filename)
        return os.path.exists(filepath)

    def __contains__(self, filename):
        """Magic method to allow constructs similar to:

            if 'abc.pkg' in dp:
        """
        return self.exists(filename)


class LocalRepository(FileRepository):
    """JAMF Pro repo located on a local filesystem path."""
    required_attrs = {"mount_point", "share_name"}

    def __init__(self, **connection_args):
        """Set up Local file share.

        Args:
            connection_args: Dict with the following key/val pairs:
                mount_point: Path to a valid mount point.
                share_name: The fileshare's name.
        """

        super(LocalRepository, self).__init__(**connection_args)
        self.connection["url"] = "local://%s" % self.connection["mount_point"]


class MountedRepository(FileRepository):
    """Parent class for mountable file shares.

    Attributes:
        fs_type: Class attribute, string protocol type (currently AFP
            or SMB).
    """
    fs_type = "undefined"

    def __init__(self, **connection_args):
        """Init a MountedRepository by calling super."""
        super(MountedRepository, self).__init__(**connection_args)

    def mount(self):
        """Mount the repository."""
        if not self.is_mounted():
            # OS X mounting is handled automagically in /Volumes:
            # DO NOT mkdir there!
            # For Linux, ensure the mountpoint exists.
            if not is_osx():
                if not os.path.exists(self.connection["mount_point"]):
                    os.mkdir(self.connection["mount_point"])
            self._mount()

    def _mount(self):
        """Private mount method."""
        raise NotImplementedError

    def umount(self, forced=True):
        """Try to unmount our mount point.

        Defaults to using forced method. If OS is Linux, it will not
        delete the mount point.

        Args:
            forced: Bool whether to force the unmount. Default is True.
        """
        if self.is_mounted():
            if is_osx():
                cmd = ["/usr/sbin/diskutil", "unmount",
                       self.connection["mount_point"]]
                if forced:
                    cmd.insert(2, "force")
                subprocess.check_call(cmd)
            else:
                cmd = ["umount", self.connection["mount_point"]]
                if forced:
                    cmd.insert(1, "-f")
                subprocess.check_call(cmd)

    def is_mounted(self):
        """Test for whether a mount point is mounted.

        If it is currently mounted, determine the path where it's
        mounted and update the connection's mount_point accordingly.
        """
        mount_check = subprocess.check_output("mount").decode().splitlines()
        # The mount command returns lines like this on OS X...
        # //username@pretendco.com/JSS%20REPO on /Volumes/JSS REPO
        # (afpfs, nodev, nosuid, mounted by local_me)
        # and like this on Linux...
        # //pretendco.com/jamf on /mnt/jamf type cifs (rw,relatime,
        # <options>...)

        valid_mount_strings = self._get_valid_mount_strings()
        was_mounted = False
        if is_osx():
            mount_string_regex = re.compile(r"\(([\w]*),*.*\)$")
            mount_point_regex = re.compile(r"on ([\w/ -]*) \(.*$")
        elif is_linux():
            mount_string_regex = re.compile(r"type ([\w]*) \(.*\)$")
            mount_point_regex = re.compile(r"on ([\w/ -]*) type .*$")
        else:
            raise JSSError("Unsupported OS.")

        for mount in mount_check:
            fs_match = re.search(mount_string_regex, mount)
            fs_type = fs_match.group(1) if fs_match else None
            # Automounts, non-network shares, and network shares
            # all have a slightly different format, so it's easiest to
            # just split.
            mount_string = mount.split(" on ")[0]
            # Does the mount_string match one of our valid_mount_strings?
            if [mstring for mstring in valid_mount_strings if
                    mstring in mount_string] and self.fs_type == fs_type:
                # Get the mount point string between from the end back to
                # the last "on", but before the options (wrapped in
                # parenthesis). Considers alphanumerics, / , _ , - and a
                # blank space as valid, but no crazy chars.
                match = re.search(mount_point_regex, mount)
                mount_point = match.group(1) if match else None
                was_mounted = True
                # Reset the connection's mount point to the discovered
                # value.
                if mount_point:
                    self.connection["mount_point"] = mount_point
                    if self.connection["jss"].verbose:
                        print ("%s is already mounted at %s.\n" %
                               (self.connection["url"], mount_point))

                # We found the share, no need to continue.
                break

        if not was_mounted:
            # If the share is not mounted, check for another share
            # mounted to the same path and if found, incremement the
            # name to avoid conflicts.
            count = 1
            while os.path.ismount(self.connection["mount_point"]):
                self.connection["mount_point"] = (
                    "%s-%s" % (self.connection["mount_point"], count))
                count += 1

        # Do an inexpensive double check...
        return os.path.ismount(self.connection["mount_point"])

    def _get_valid_mount_strings(self):
        """Return a tuple of potential mount strings.

        Casper Admin seems to mount in a number of ways:
            - hostname/share
            - fqdn/share
        Plus, there's the possibility of:
            - IPAddress/share
        Then factor in the possibility that the port is included too!
        This gives us a total of up to six valid addresses for mount
        to report.
        """
        results = set()
        join = os.path.join
        url = self.connection["url"]
        share_name = quote(self.connection["share_name"],
                                  safe="~()*!.'$")
        port = self.connection["port"]

        # URL from python-jss form:
        results.add(join(url, share_name))
        results.add(join("%s:%s" % (url, port), share_name))

        # IP Address form:
        # socket.gethostbyname() will return an IP address whether
        # an IP address, FQDN, or .local name is provided.
        ip_address = socket.gethostbyname(url)
        results.add(join(ip_address, share_name))
        results.add(join("%s:%s" % (ip_address, port), share_name))

        # Domain name only form:
        domain_name = url.split(".")[0]
        results.add(join(domain_name, share_name))
        results.add(join("%s:%s" % (domain_name, port), share_name))

        # FQDN form using getfqdn:
        # socket.getfqdn() could just resolve back to the ip
        # or be the same as the initial URL so only add it if it's
        # different than both.
        fqdn = socket.getfqdn(ip_address)
        results.add(join(fqdn, share_name))
        results.add(join("%s:%s" % (fqdn, port), share_name))

        return tuple(results)

    @auto_mounter
    def _copy(self, filename, destination):
        """Copy a file or folder to the repository.

        Will mount if needed.

        Args:
            filename: Path to copy.
            destination: Remote path to copy file to.
        """
        super(MountedRepository, self)._copy(filename, destination)

    @auto_mounter
    def delete(self, filename):
        """Delete a file from the repository.

        Args:
            filename: String filename only (i.e. no path) of file to
                delete.
        """
        super(MountedRepository, self).delete(filename)

    @auto_mounter
    def exists(self, filename):
        """Report whether a file exists on the distribution point.

        Determines file type by extension.

        Args:
            filename: Filename you wish to check. (No path! e.g.:
                "AdobeFlashPlayer-14.0.0.176.pkg")
        """
        return super(MountedRepository, self).exists(filename)

    def __repr__(self):
        """Return a formatted string of connection info."""
        # Do an "update" to get current mount points.
        self.is_mounted()
        output = super(MountedRepository, self).__repr__()
        output += "Mounted: %s\n" % self.is_mounted()
        return output

    @property
    def _encoded_password(self):
        """Returns the safely url-quoted password for this DP."""
        return quote(self.connection["password"], safe="~()*!.'$")


class AFPDistributionPoint(MountedRepository):
    """Represents an AFP repository."""
    protocol = "afp"
    fs_type = "afpfs"
    required_attrs = {"url", "mount_point", "username", "password",
                      "share_name"}

    def __init__(self, **connection_args):
        """Set up an AFP connection.

        Args:
            connection_args (dict): Dict with the following key/val pairs:
                url: URL to the mountpoint,including volume name e.g.:
                    "my_repository.domain.org/jamf" (Do _not_ include
                    protocol or auth info.)
                mount_point: Path to a valid mount point.
                share_name: The fileshare's name.
                username: Share R/W username.
                password: Share R/W password.
        """
        super(AFPDistributionPoint, self).__init__(**connection_args)
        # Check to see if share is mounted, and update mount point
        self.is_mounted()

    def _build_url(self):
        """Build the URL string to mount this file share."""
        if self.connection.get("username") and self.connection.get("password"):
            auth = "%s:%s@" % (self.connection["username"],
                               self._encoded_password)
        else:
            auth = ""

        # Optional port number
        port = self.connection.get("port")
        port = ":" + port if port else ""

        self.connection["mount_url"] = "%s://%s%s%s/%s" % (
            self.protocol, auth, self.connection["url"], port,
            self.connection["share_name"])

    def _mount(self):
        """Mount based on which OS is running."""
        # mount_afp "afp://scraig:<password>@address/share" <mnt_point>
        if is_osx():
            if self.connection["jss"].verbose:
                print(self.connection["mount_url"])
            if mount_share:
                self.connection["mount_point"] = mount_share(
                    self.connection["mount_url"])
            else:
                # Non-Apple OS X python:
                args = ["mount", "-t", self.protocol,
                        self.connection["mount_url"],
                        self.connection["mount_point"]]
                if self.connection["jss"].verbose:
                    print(" ".join(args))
                subprocess.check_call(args)
        elif is_linux():
            args = ["mount_afp", "-t", self.protocol,
                    self.connection["mount_url"],
                    self.connection["mount_point"]]
            if self.connection["jss"].verbose:
                print(" ".join(args))
            subprocess.check_call(args)
        else:
            raise JSSError("Unsupported OS.")


class SMBDistributionPoint(MountedRepository):
    """Represents a SMB distribution point."""
    protocol = "smbfs"
    required_attrs = {"url", "share_name", "mount_point", "domain", "username",
                      "password"}

    def __init__(self, **connection_args):
        """Set up a SMB connection.

        Args:
            connection_args: Dict with the following key/val pairs:
                url: URL to the mountpoint,including volume name e.g.:
                    "my_repository.domain.org/jamf" (Do _not_ include
                    protocol or auth info.)
                mount_point: Path to a valid mount point.
                share_name: The fileshare's name.
                domain: Specify the domain.
                username: Share R/W username.
                password: Share R/W password.
        """
        super(SMBDistributionPoint, self).__init__(**connection_args)
        if is_osx():
            self.fs_type = "smbfs"
        if is_linux():
            self.fs_type = "cifs"
        # Check to see if share is mounted, and update.
        self.is_mounted()

    def _build_url(self):
        """Build the URL string to mount this file share."""
        if self.connection.get("username") and self.connection.get("password"):
            auth = "%s:%s@" % (self.connection["username"],
                               self._encoded_password)
            if self.connection.get("domain"):
                auth = r"%s;%s" % (self.connection["domain"], auth)
        else:
            auth = ""

        # Optional port number
        port = self.connection.get("port")
        port = ":" + port if port else ""

        # Construct mount_url
        self.connection["mount_url"] = "//%s%s%s/%s" % (
            auth, self.connection["url"], port, self.connection["share_name"])

    def _mount(self):
        """Mount based on which OS is running."""
        # mount -t cifs -o \
        # username=<user>,password=<password>,domain=<domain>,port=445 \
        # //server/share /mnt/<mountpoint>
        if is_osx():
            if mount_share:
                mount_url = "smb:%s" % self.connection["mount_url"]
                if self.connection["jss"].verbose:
                    print(mount_url)
                self.connection["mount_point"] = mount_share(mount_url)
            else:
                # Non-Apple OS X python:
                args = ["mount", "-t", self.protocol,
                        self.connection["mount_url"],
                        self.connection["mount_point"]]
                if self.connection["jss"].verbose:
                    print(" ".join(args))
                subprocess.check_call(args)
        elif is_linux():
            args = ["mount", "-t", "cifs", "-o",
                    "username=%s,password=%s,domain=%s,port=%s" %
                    (self.connection["username"], self.connection["password"],
                     self.connection["domain"], self.connection["port"]),
                    "//%s/%s" % (self.connection["url"],
                                 self.connection["share_name"]),
                    self.connection["mount_point"]]
            if self.connection["jss"].verbose:
                print(" ".join(args))
            subprocess.check_call(args)
        else:
            raise JSSError("Unsupported OS.")


class DistributionServer(Repository):
    """Abstract class for representing JDS and CDP type repos.

    The JSS has a folder to which packages are uploaded via a private
    API call to dbfileupload. From there, the JSS handles the
    distribution to its Cloud and JDS points.

    There are caveats to its exists() method which you should be
    aware of, along with a private API exists_with_casper method, which
    probably works more like what one would expect. Please see those
    methods for more information.
    """
    required_attrs = {"jss"}
    destination = "0"

    def __init__(self, **connection_args):
        """Set up a connection to a distribution server.

        Args:
            connection_args: Dict, with required key:
                jss: A JSS Object.
        """
        super(DistributionServer, self).__init__(**connection_args)
        self.connection["url"] = self.connection["jss"].base_url

    def _build_url(self):
        """Build the URL for POSTing files. 10.2 and earlier.

        This actually still works in some scenarios, but it seems like it will be deprecated soon.
        """
        self.connection["upload_url"] = (
                "%s/%s" % (self.connection["jss"].base_url, "dbfileupload"))
        self.connection["delete_url"] = (
                "%s/%s" % (self.connection["jss"].base_url,
                           "casperAdminSave.jxml"))

    def _build_url_modern(self):
        """Build the URL for POSTing files.

        This uses the UploadServlet that has been used to handle most file uploads into JAMF Pro.
        """
        self.connection["upload_url"] = (
            "%s/%s" % (self.connection["jss"].base_url, "upload"))
        self.connection["delete_url"] = (
            "%s/%s" % (self.connection["jss"].base_url,
                       "casperAdminSave.jxml"))

    def copy_pkg(self, filename, id_=-1):
        """Copy a package to the distribution server.

        Bundle-style packages must be zipped prior to copying.

        Args:
            filename: Full path to file to upload.
            id_: ID of Package object to associate with, or -1 for new
                packages (default).
        """
        self._copy(filename, id_=id_, file_type=PKG_FILE_TYPE)

    def _copy(self, filename, id_=-1, file_type=0):
        """Upload a file to the distribution server. 10.2 and earlier

        Directories/bundle-style packages must be zipped prior to
        copying.
        """
        if os.path.isdir(filename):
            raise TypeError(
                "Distribution Server type repos do not permit directory "
                "uploads. You are probably trying to upload a non-flat "
                "package. Please zip or create a flat package.")
        basefname = os.path.basename(filename)
        resource = open(filename, "rb")
        headers = {"DESTINATION": self.destination, "OBJECT_ID": str(id_),
                   "FILE_TYPE": file_type, "FILE_NAME": basefname}
        response = self.connection["jss"].session.post(
            url=self.connection["upload_url"],
            data=resource.read(),
            headers=headers)
        if self.connection["jss"].verbose:
            print(response)

    def _copy_new(self, filename, id_=-1, file_type=0):
        """Upload a file to the distribution server.

        Directories/bundle-style packages must be zipped prior to
        copying.
        """
        if os.path.isdir(filename):
            raise TypeError(
                "Distribution Server type repos do not permit directory "
                "uploads. You are probably trying to upload a non-flat "
                "package. Please zip or create a flat package.")
        basefname = os.path.basename(filename)
        resource = open(filename, "rb")
        headers = {"sessionIdentifier": "com.jamfsoftware.jss.objects.packages.Package:%s" % str(id_),
                   "fileIdentifier": "FIELD_FILE_NAME_FOR_DIST_POINTS"}
        response = self.connection["jss"].session.post(
            url=self.connection["upload_url"],
            data=resource.read(),
            headers=headers)
        print(response)
        if self.connection["jss"].verbose:
            print(response)

    def delete_with_casper_admin_save(self, pkg):
        """Delete a pkg from the distribution server.

        Args:
            pkg: Can be a jss.Package object, an int ID of a package, or
                a filename.
        """
        # The POST needs the package ID.
        if pkg.__class__.__name__ == "Package":
            package_to_delete = pkg.id
        elif isinstance(pkg, int):
            package_to_delete = pkg
        elif isinstance(pkg, str):
            package_to_delete = self.connection["jss"].Package(pkg).id
        else:
            raise TypeError

        data_dict = {"username": self.connection["jss"].user,
                     "password": self.connection["jss"].password,
                     "deletedPackageID": package_to_delete}
        self.connection["jss"].session.post(
            url=self.connection["delete_url"], data=data_dict)
        # There's no response if it works.

    def delete(self, filename):
        """Delete a package distribution server.

        This method simply finds the Package object from the database
        with the API GET call and then deletes it. This will remove the
        file from the database blob.

        For setups which have file share distribution points, you will
        need to delete the files on the shares also.

        Args:
            filename: Filename (no path) to delete.
        """
        if is_package(filename):
            self.connection["jss"].Package(filename).delete()

    def exists(self, filename):
        """Check for the existence of a package.

        Unlike other DistributionPoint types, JDS and CDP types have no
        documented interface for checking whether the server and its
        children have a complete copy of a file. The best we can do is
        check for an object using the API /packages URL--JSS.Package()
        and look for matches on the filename.

        If this is not enough, please use the alternate
        exists_with_casper method.  For example, it's possible to create
        a Package object but never upload a package file, and this
        method will still return "True".

        Also, this may be slow, as it needs to retrieve the complete
        list of packages from the server.
        """
        # Technically, the results of the casper.jxml page list the
        # package files on the server. This is an undocumented
        # interface, however.
        result = False
        if is_package(filename):
            packages = self.connection["jss"].Package().retrieve_all()
            for package in packages:
                if package.findtext("filename") == filename:
                    result = True
                    break

        return result

    def exists_using_casper(self, filename):
        """Check for the existence of a package file.

        Unlike other DistributionPoint types, JDS and CDP types have no
        documented interface for checking whether the server and its
        children have a complete copy of a file. The best we can do is
        check for an object using the API /packages URL--JSS.Package()
        and look for matches on the filename.

        If this is not enough, this method uses the results of the
        casper.jxml page to determine if a package exists. This is an
        undocumented feature and as such should probably not be relied
        upon.

        It will test for whether the file exists on ALL configured
        distribution servers. This may register False if the JDS is busy
        syncing them.
        """
        casper_results = casper.Casper(self.connection["jss"])
        distribution_servers = casper_results.find("distributionservers")

        # Step one: Build a list of sets of all package names.
        all_packages = []
        for distribution_server in distribution_servers:
            packages = set()
            for package in distribution_server.findall("packages/package"):
                packages.add(os.path.basename(package.find("fileURL").text))

            all_packages.append(packages)

        # Step two: Intersect the sets.
        base_set = all_packages.pop()
        for packages in all_packages:
            base_set = base_set.intersection(packages)

        # Step three: Check for membership.
        return filename in base_set


class JDS(DistributionServer):
    """Class for representing a JDS and its controlling JSS.

    The JSS has a folder to which packages are uploaded. From there, the
    JSS handles the distribution to its Cloud and JDS points.

    This class should be considered experimental!
    - There are caveats to its .exists() method
    - It is unclear at the moment what the interaction is in systems
    that have both a JDS and a CDP, especially depending on which is the
    master.
    """
    required_attrs = {"jss"}
    destination = "1"


class CDP(DistributionServer):
    """Class for representing a CDP and its controlling JSS.

    The JSS has a folder to which packages are uploaded. From there, the
    JSS handles the distribution to its Cloud and JDS points.

    This class should be considered experimental!
    - There are caveats to its .exists() method
    - It is unclear at the moment what the interaction is in systems
    that have both a JDS and a CDP, especially depending on which is the
    master.
    """
    required_attrs = {"jss"}
    destination = "2"


class CloudDistributionServer(Repository):
    """Abstract class for representing JCDS type repos.

    """
    def package_index_using_casper(self):
        """Get a list of packages on the JCDS

        Similar to JDS and CDP, JCDS types have no
        documented interface for checking whether the server and its
        children have a complete copy of a file. The best we can do is
        check for an object using the API /packages URL--JSS.Package()
        and look for matches on the filename.

        If this is not enough, this method uses the results of the
        casper.jxml page to determine if a package exists. This is an
        undocumented feature and as such should probably not be relied
        upon.

        It will test for whether the file exists on only cloud distribution points.
        """
        casper_results = casper.Casper(self.connection["jss"])
        cloud_distribution_points = casper_results.find("cloudDistributionPoints")

        # Step one: Build a list of sets of all package names.
        all_packages = []
        for distribution_point in cloud_distribution_points:
            if distribution_point.findtext('name') != 'Jamf Cloud':
                continue  # type 4 might be reserved for JCDS?

            for package in distribution_point.findall("packages/package"):
                package_obj = casper_results.find("./packages/package[id='%s']" % (package.findtext('id'),))

                all_packages.append({
                    'id': package.findtext('id'),
                    'checksum': package.findtext('checksum'),
                    'size': package.findtext('size'),
                    'lastModified': package.findtext('lastModified'),
                    'fileURL': unquote(package.findtext('fileURL')),
                    'name': package_obj.findtext('name'),
                    'filename': package_obj.findtext('filename'),
                })

        return all_packages


def _jcds_upload_chunk(
        filename,
        base_url,
        upload_token,
        chunk_index,
        chunk_size,
        total_chunks):
    """Upload a single chunk of a file to JCDS.

    Args:
        filename (str): The full path to the file being uploaded.
        base_url (str): The JCDS base URL which includes the regional hostname and the tenant id.
        upload_token (str): The upload token, scraped from legacy/packages.html
        chunk_index (int): The zero-based index of the chunk being uploaded.
        total_chunks (int): The total count of chunks to upload

    Returns:
        dict: JSON Response from JCDS
    """
    print("Working on Chunk [{}/{}]".format(chunk_index + 1, total_chunks))
    resource = open(filename, "rb")
    resource.seek(chunk_index * chunk_size)
    chunk_data = resource.read(chunk_size)
    basefname = os.path.basename(filename)
    chunk_url = "{}/{}/part?chunk={}&chunks={}".format(
        base_url, basefname, chunk_index, total_chunks
    )

    chunk_reader = io.BytesIO(chunk_data)
    headers = {"X-Auth-Token": upload_token}
    response = requests.post(
        url=chunk_url,
        headers=headers,
        files={'file': chunk_reader},
    )

    return response.json()


# Semaphore controlling max workers for chunked uploads
jcds_semaphore = threading.BoundedSemaphore(value=3)


class JCDSChunkUploadThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        self.filename = kwargs['filename']
        self.base_url = kwargs['base_url']
        self.upload_token = kwargs['upload_token']
        self.chunk_index = kwargs['chunk_index']
        self.chunk_size = kwargs['chunk_size']
        self.total_chunks = kwargs['total_chunks']

        super_kwargs = dict(kwargs)
        del super_kwargs['filename']
        del super_kwargs['base_url']
        del super_kwargs['upload_token']
        del super_kwargs['chunk_index']
        del super_kwargs['chunk_size']
        del super_kwargs['total_chunks']

        super(JCDSChunkUploadThread, self).__init__(*args, **super_kwargs)

    def run(self):
        jcds_semaphore.acquire()
        try:
            print("Working on Chunk [{}/{}]".format(self.chunk_index + 1, self.total_chunks))

            resource = open(self.filename, "rb")
            resource.seek(self.chunk_index * self.chunk_size)
            chunk_data = resource.read(self.chunk_size)
            basefname = os.path.basename(self.filename)
            chunk_url = "{}/{}/part?chunk={}&chunks={}".format(
                self.base_url, basefname, self.chunk_index, self.total_chunks
            )

            chunk_reader = io.BytesIO(chunk_data)
            headers = {"X-Auth-Token": self.upload_token}
            response = requests.post(
                url=chunk_url,
                headers=headers,
                files={'file': chunk_reader},
            )

            return response.json()
        except:
            pass
        finally:
            jcds_semaphore.release()


class AWS(CloudDistributionServer, abstract.AbstractRepository):
    """Class for representing an AWS Cloud Distribution Point and its controlling JSS.

    """
    required_attrs = {"jss", "bucket"}

    def __init__(self, **connection_args):
        """Set up a connection to an AWS S3 bucket.

        It is more secure to use the following environment variables provided by boto:

            AWS_ACCESS_KEY_ID - The access key id to the jamf bucket
            AWS_SECRET_ACCESS_KEY - The secret access key to the jamf bucket

        You may also use the file ~/.boto as described in the boto documentation.

        Args:
            connection_args: Dict, with required keys:
                jss: A JSS Object.
                bucket: Name of the JAMF bucket.
                aws_access_key_id (optional): The access key id
                secret_access_key (optional): The secret access key, use environment instead.
                host (optional): A bucket host. Seems to be needed if your bucket is not in the default location
                    eg. southeast asia ap 2
                chunk_size (optional): The chunk size for large objects >50mb

        Throws:
            S3ResponseError if the bucket does not exist
        """
        super(AWS, self).__init__(**connection_args)
        self.s3 = S3Connection(
            aws_access_key_id=connection_args.get('aws_access_key_id', None),
            aws_secret_access_key=connection_args.get('aws_secret_access_key', None),
            host=connection_args.get('host', boto.s3.connection.NoHostProvided),
        )
        try:
            self.bucket = self.s3.get_bucket(connection_args['bucket'])
        except S3ResponseError as e:
            raise JSSError("got error getting bucket, may not exist: {}".format(connection_args['bucket']))

        self.connection["url"] = self.bucket
        self.chunk_size = connection_args.get('chunk_size', 52428800)  # 50 mb default

    def _build_url(self):
        """Build a connection URL."""
        pass

    def copy_pkg(self, filename, id_=-1):
        """Copy a package to the repo's Package subdirectory.

        Args:
            filename: Path for file to copy.
            id_: Unused
        """
        self._copy(filename, id_=id_)

    def _copy(self, filename, id_=-1):   # type: (str, int) -> None
        """Copy a file or folder to the bucket.

        Does not yet support chunking.

        Args:
            filename: Path to copy.
            destination: Remote path to copy file to.
        """
        bucket_key = os.path.basename(filename)
        exists = self.bucket.get_key(bucket_key)
        if exists:
            print("Already exists")
        else:
            k = Key(self.bucket)
            k.key = bucket_key
            k.set_metadata('jamf-package-id', id_)
            k.set_contents_from_filename(filename)

    def delete(self, filename):  # type: (str) -> None
        bucket_key = os.path.basename(filename)
        self.bucket.delete_key(bucket_key)

    def exists(self, filename):  # type: (str) -> bool
        """Check whether a package already exists by checking for a bucket item with the same filename.

        Args:
            filename: full path to filename. Only the name itself will be checked.

        Returns:
            True if the package exists, else false
        """
        k = self.bucket.get_key(os.path.basename(filename))
        return k is not None


class JCDS(CloudDistributionServer):
    """Class for representing a JCDS and its controlling jamfcloud JSS.

    The JSS allows direct upload to the JCDS by exposing the access token from the package upload page.

    This class should be considered experimental!
    """
    required_attrs = {"jss"}
    destination = "3"
    workers = 3
    chunk_size = 1048768

    def __init__(self, **connection_args):
        """Set up a connection to a distribution server.

        Args:
            connection_args (dict):
                jss (JSS): The associated JAMF Pro Server instance
        """
        super(JCDS, self).__init__(**connection_args)
        self.connection["url"] = "JCDS"

    def _scrape_tokens(self):
        """Scrape JCDS upload URL and upload access token from the jamfcloud instance."""
        jss = self.connection['jss']
        response = jss.scrape('legacy/packages.html?id=-1&o=c')
        matches = re.search(r'data-base-url="([^"]*)"', response.content.decode("utf-8"))
        if matches is None:
            raise JSSError('Did not find the JCDS base URL on the packages page. Is this actually Jamfcloud?')

        jcds_base_url = matches.group(1)

        matches = re.search(r'data-upload-token="([^"]*)"', response.content.decode("utf-8"))
        if matches is None:
            raise JSSError('Did not find the JCDS upload token on the packages page. Is this actually Jamfcloud?')

        jcds_upload_token = matches.group(1)

        h = HTMLParser()
        jcds_base_url = h.unescape(jcds_base_url)
        self.connection['jcds_base_url'] = jcds_base_url
        self.connection['jcds_upload_token'] = jcds_upload_token
        self.connection["url"] = jcds_base_url  # This is to make JSSImporter happy because it accesses .connection

    def _build_url(self):
        """Build a connection URL."""
        pass

    def copy_pkg(self, filename, id_=-1):
        """Copy a package to the JAMF Cloud distribution server.

        Bundle-style packages must be zipped prior to copying.

        Args:
            filename: Full path to file to upload.
            id_: ID of Package object to associate with, or -1 for new
                packages (default).
        """
        self._copy(filename, id_=id_)

    def _build_chunk_url(self, filename, chunk, chunk_total):
        """Build the path to the chunk being uploaded to the JCDS."""
        return "{}/{}/part?chunk={}&chunks={}".format(
            self.connection["jcds_base_url"], filename, chunk, chunk_total
        )

    def _copy_multiprocess(self, filename, upload_token, id_=-1):
        """Upload a file to the distribution server using multiple processes to upload several chunks in parallel.

        Directories/bundle-style packages must be zipped prior to copying.
        """
        fsize = os.stat(filename).st_size
        total_chunks = int(math.ceil(fsize / JCDS.chunk_size))
        p = multiprocessing.Pool(3)

        def _chunk_args(chunk_index):
            return [filename, self.connection["jcds_base_url"], upload_token, chunk_index, JCDS.chunk_size, total_chunks]

        for chunk in xrange(0, total_chunks):
            res = p.apply_async(_jcds_upload_chunk, _chunk_args(chunk))
            data = res.get(timeout=10)
            print("id: {0}, version: {1}, size: {2}, filename: {3}, lastModified: {4}, created: {5}".format(
                data['id'], data['version'], data['size'], data['filename'], data['lastModified'], data['created']))

    def _copy_threaded(self, filename, upload_token, id_=-1):
        """Upload a file to the distribution server using multiple threads to upload several chunks in parallel."""
        fsize = os.stat(filename).st_size
        total_chunks = int(math.ceil(fsize / JCDS.chunk_size))

        for chunk in xrange(0, total_chunks):
            t = JCDSChunkUploadThread(
                filename=filename,
                base_url=self.connection["jcds_base_url"],
                upload_token=upload_token,
                chunk_index=chunk,
                chunk_size=JCDS.chunk_size,
                total_chunks=total_chunks,
            )
            t.start()

    def _copy_sequential(self, filename, upload_token, id_=-1):
        """Upload a file to the distribution server using the same process as python-jss.

        Directories/bundle-style packages must be zipped prior to copying.
        """
        fsize = os.stat(filename).st_size
        total_chunks = int(math.ceil(fsize / JCDS.chunk_size))

        basefname = os.path.basename(filename)
        resource = open(filename, "rb")

        headers = {
            "X-Auth-Token": self.connection['jcds_upload_token'],
            # "Content-Type": "application/octet-steam",
        }

        for chunk in xrange(0, total_chunks):
            resource.seek(chunk * JCDS.chunk_size)
            chunk_data = resource.read(JCDS.chunk_size)
            chunk_reader = io.BytesIO(chunk_data)
            chunk_url = self._build_chunk_url(basefname, chunk, total_chunks)
            response = self.connection["jss"].session.post(
                url=chunk_url,
                headers=headers,
                files={'file': chunk_reader},
            )

            if self.connection["jss"].verbose:
                print(response.json())

        resource.close()

    def _copy(self, filename, id_=-1, file_type=0):
        """Upload a file to the distribution server. 10.2 and earlier

        Directories/bundle-style packages must be zipped prior to
        copying.

        JCDS returns a JSON structure like this::

            {
                u'status': u'PENDING',
                u'created': u'2018-07-10T03:21:17.000Z',
                u'lastModified': u'2018-07-11T03:55:32.000Z',
                u'filename': u'SkypeForBusinessInstaller-16.18.0.51.pkg',
                u'version': 6,
                u'md5': u'',
                u'sha512': u'',
                u'id': u'3a7e6a7479fc4000bf53a9693d906b11',
                u'size': 35658112
            }

        """
        if os.path.isdir(filename):
            raise TypeError(
                "JCDS Server type repos do not permit directory "
                "uploads. You are probably trying to upload a non-flat "
                "package. Please zip or create a flat package.")

        if 'jcds_upload_token' not in self.connection:
            self._scrape_tokens()

        self._copy_threaded(filename, self.connection['jcds_upload_token'])
        # if False:
        #self._copy_sequential(filename, self.connection['jcds_upload_token'])
        # else:
        #     self._copy_threaded(filename, self.connection['jcds_upload_token'])

    def exists(self, filename):
        """Check whether a package file already exists."""
        packages = self.package_index_using_casper()
        for p in packages:
            url, token = p['fileURL'].split('?', 2)
            urlparts = url.split('/')

            if urlparts[-1] == filename:
                return True

        return False

    def __repr__(self):
        """Return string representation of connection arguments."""
        output = ["JAMF Cloud Distribution Server: %s" % self.connection["jss"].base_url]
        output.append("Type: %s" % type(self))
        output.append("Connection Information:")
        for key, val in self.connection.items():
            output.append("\t%s: %s" % (key, val))

        return "\n".join(output) + "\n"



"""nsurlsession_adapter.py

Requests API transport that uses NSURLSession and friends via PyObjC 2.1 that ships with macOS.

Much of this code was copied, rewritten, or inspired by gurl.py from the munki project.
Thanks to Greg Neagle and frogor, amongst many other contributors.
"""
from __future__ import absolute_import
import logging
import os
import io
from requests.adapters import BaseAdapter
from requests.models import Response, PreparedRequest, Request
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers
from requests.auth import AuthBase, HTTPBasicAuth
from requests.cookies import RequestsCookieJar
from requests.exceptions import SSLError
# from cookielib import Cookie

import objc
from Foundation import NSObject, NSMutableURLRequest, NSURL, NSURLRequestUseProtocolCachePolicy, \
    NSURLRequestReloadIgnoringLocalCacheData, NSURLRequestReturnCacheDataElseLoad, \
    NSURLSessionConfiguration, NSURLSession, NSOperationQueue, NSURLCredential, \
    NSURLCredentialPersistenceNone, NSRunLoop, NSDate, NSURLResponseUnknownLength, \
    NSURLAuthenticationMethodServerTrust, NSMutableData, NSBundle


# These headers are reserved to the NSURLSession and should not be set by
# Requests, though we may use them.
_RESERVED_HEADERS = set([
    'connection', 'host', 'proxy-authenticate', 'authorization',
    'proxy-authorization', 'www-authenticate',
])

# NSURLSessionAuthChallengeDisposition enum constants
NSURLSessionAuthChallengeUseCredential = 0
NSURLSessionAuthChallengePerformDefaultHandling = 1
NSURLSessionAuthChallengeCancelAuthenticationChallenge = 2
NSURLSessionAuthChallengeRejectProtectionSpace = 3

# NSURLSessionResponseDisposition enum constants
NSURLSessionResponseCancel = 0
NSURLSessionResponseAllow = 1
NSURLSessionResponseBecomeDownload = 2

# TLS/SSLProtocol enum constants
kSSLProtocolUnknown = 0
kSSLProtocol3 = 2
kTLSProtocol1 = 4
kTLSProtocol11 = 7
kTLSProtocol12 = 8
kDTLSProtocol1 = 9

ssl_error_codes = {
    -9800: u'SSL protocol error',
    -9801: u'Cipher Suite negotiation failure',
    -9802: u'Fatal alert',
    -9803: u'I/O would block (not fatal)',
    -9804: u'Attempt to restore an unknown session',
    -9805: u'Connection closed gracefully',
    -9806: u'Connection closed via error',
    -9807: u'Invalid certificate chain',
    -9808: u'Bad certificate format',
    -9809: u'Underlying cryptographic error',
    -9810: u'Internal error',
    -9811: u'Module attach failure',
    -9812: u'Valid cert chain, untrusted root',
    -9813: u'Cert chain not verified by root',
    -9814: u'Chain had an expired cert',
    -9815: u'Chain had a cert not yet valid',
    -9816: u'Server closed session with no notification',
    -9817: u'Insufficient buffer provided',
    -9818: u'Bad SSLCipherSuite',
    -9819: u'Unexpected message received',
    -9820: u'Bad MAC',
    -9821: u'Decryption failed',
    -9822: u'Record overflow',
    -9823: u'Decompression failure',
    -9824: u'Handshake failure',
    -9825: u'Misc. bad certificate',
    -9826: u'Bad unsupported cert format',
    -9827: u'Certificate revoked',
    -9828: u'Certificate expired',
    -9829: u'Unknown certificate',
    -9830: u'Illegal parameter',
    -9831: u'Unknown Cert Authority',
    -9832: u'Access denied',
    -9833: u'Decoding error',
    -9834: u'Decryption error',
    -9835: u'Export restriction',
    -9836: u'Bad protocol version',
    -9837: u'Insufficient security',
    -9838: u'Internal error',
    -9839: u'User canceled',
    -9840: u'No renegotiation allowed',
    -9841: u'Peer cert is valid, or was ignored if verification disabled',
    -9842: u'Server has requested a client cert',
    -9843: u'Peer host name mismatch',
    -9844: u'Peer dropped connection before responding',
    -9845: u'Decryption failure',
    -9846: u'Bad MAC',
    -9847: u'Record overflow',
    -9848: u'Configuration error',
    -9849: u'Unexpected (skipped) record in DTLS'}

# define a helper function for block callbacks
import ctypes
import objc
_objc_so = ctypes.cdll.LoadLibrary(
    os.path.join(objc.__path__[0], '_objc.so'))
PyObjCMethodSignature_WithMetaData = (
    _objc_so.PyObjCMethodSignature_WithMetaData)
PyObjCMethodSignature_WithMetaData.restype = ctypes.py_object


def objc_method_signature(signature_str):
    """Return a PyObjCMethodSignature given a call signature in string
    format"""
    return PyObjCMethodSignature_WithMetaData(
        ctypes.create_string_buffer(signature_str), None, False)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

# ATS hack stolen from gurl.py
# disturbing hack warning!
# this works around an issue with App Transport Security on 10.11
bundle = NSBundle.mainBundle()
info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
info['NSAppTransportSecurity'] = {'NSAllowsArbitraryLoads': True}


def build_request(request, timeout):
    # type: (PreparedRequest, Union[float, Tuple[float, float]]) -> NSMutableURLRequest
    """Convert a requests `PreparedRequest` into a suitable NSMutableURLRequest."""

    timeout = timeout if timeout else 0.0

    nsrequest = NSMutableURLRequest.requestWithURL_cachePolicy_timeoutInterval_(
        NSURL.URLWithString_(request.url),
        NSURLRequestReloadIgnoringLocalCacheData,
        timeout,
    )
    nsrequest.setHTTPMethod_(request.method)

    for k, v in request.headers.items():
        k, v = k.lower(), v.lower()

        if k in _RESERVED_HEADERS:
            logging.debug('Ignoring reserved header: %s', k)
            continue

        nsrequest.setValue_forHTTPHeaderField_(v, k)

    return nsrequest


def build_response(req, delegate, cookiestore):
    # type: (PreparedRequest, NSURLSessionAdapterDelegate, NSHTTPCookieStorage) -> Response
    """Build a requests `Response` object from the response data collected by the NSURLSessionDelegate, and the
    default cookie store."""

    response = Response()

    # Fallback to None if there's no status_code, for whatever reason.
    response.status_code = getattr(delegate, 'status', None)

    # Make headers case-insensitive.
    response.headers = CaseInsensitiveDict(getattr(delegate, 'headers', {}))

    # Set encoding.
    response.encoding = get_encoding_from_headers(response.headers)
    # response.raw = resp
    # response.reason = response.raw.reason

    if isinstance(req.url, bytes):
        response.url = req.url.decode('utf-8')
    else:
        response.url = req.url

    # Add new cookies from the server. For NSURLSession these have already been parsed.
    # jar = RequestsCookieJar()
    # for cookie in cookiestore.cookies():
    #     print cookie
    #     c = Cookie(
    #         version=cookie.version(),
    #         name=cookie.name(),
    #         value=cookie.value(),
    #         port=8444,
    #         # port=cookie.portList()[-1],
    #         port_specified=0,
    #         domain=cookie.domain(),
    #         domain_specified=cookie.domain(),
    #         domain_initial_dot=cookie.domain(),
    #         path=cookie.path(),
    #         path_specified=cookie.path(),
    #         secure=!cookie.HTTPOnly(),
    #         expires=cookie.expiresDate(),
    #         comment=cookie.comment(),
    #         comment_url=cookie.commentURL(),
    #     )
    #     jar.set_cookie(c)
    #
    # response.cookies = jar

    response.raw = io.BytesIO(buffer(delegate.output))

    # Give the Response some context.
    response.request = req
    # response.connection = self

    return response


class NSURLCredentialAuth(AuthBase):
    """HTTP Basic Authentication for NSURLSessionAdapter.

    It isn't possible to use the requests built-in HTTPBasicAuth adapter because it adds headers, which doesn't really
    gel with NSURLSession's delegate challenge methods."""

    def __init__(self, username, password):  # type: (str, str) -> None
        self.username = username
        self.password = password
        self.credential = None

    def __eq__(self, other):  # type: (Any) -> bool
        return all([
            self.username == getattr(other, 'username', None),
            self.password == getattr(other, 'password', None)
        ])

    def __ne__(self, other):  # type: (Any) -> bool
        return not self == other

    def __call__(self, r):  # type: (PreparedRequest) -> PreparedRequest
        """Instead of modifying the request object, we construct an instance of NSURLCredential to attach to ourselves.

        When the delegate detects that attribute is present, it uses it whenever a challenge comes in."""
        credential = NSURLCredential.credentialWithUser_password_persistence_(
            self.username, self.password,
            NSURLCredentialPersistenceNone  # we don't expect ephemeral requests to save keychain items.
        )
        self.credential = credential

        return r


class NSURLSessionAdapterDelegate(NSObject):
    """NSURLSessionAdapterDelegate implements the delegate methods of NSURLSession protocols such as:

    - `NSURLSessionDelegate <https://developer.apple.com/documentation/foundation/nsurlsessiondelegate?language=objc>`_.
    - `NSURLSessionTaskDelegate <https://developer.apple.com/documentation/foundation/nsurlsessiontaskdelegate?language=objc>`_.
    - `NSURLSessionDataDelegate <https://developer.apple.com/documentation/foundation/nsurlsessiondatadelegate?language=objc>`_.
    """

    def initWithAdapter_(self, adapter):
        self = objc.super(NSURLSessionAdapterDelegate, self).init()

        if self is None:
            return None

        self.adapter = adapter
        self.bytesReceived = 0
        self.expectedLength = -1
        self.percentComplete = 0
        self.done = False
        self.error = None
        self.SSLError = None
        self.output = NSMutableData.dataWithCapacity_(1024)
        self.status = None
        self.headers = {}
        self.verify = True
        self.credential = None
        self.history = []

        return self

    def isDone(self):  # () -> bool
        """Check if the connection request is complete. As a side effect,
        allow the delegates to work by letting the run loop run for a bit"""
        if self.done:
            return self.done
        # let the delegates do their thing
        NSRunLoop.currentRunLoop().runUntilDate_(
            NSDate.dateWithTimeIntervalSinceNow_(.1))
        return self.done

    def URLSession_dataTask_didReceiveResponse_completionHandler_(
            self,
            session,           # type: NSURLSession
            task,              # type: NSURLSessionDataTask
            response,          # type: NSHTTPURLResponse
            completionHandler  # type: (NSURLSessionResponseDisposition) -> Void
    ):  # type: (...) -> None
        logger.debug('URLSession_dataTask_didReceiveResponse_completionHandler_')
        completionHandler.__block_signature__ = objc_method_signature('v@i')
        
        self.response = response
        self.bytesReceived = 0
        self.percentComplete = -1
        self.expectedLength = response.expectedContentLength()

        if response.className() == u'NSHTTPURLResponse':
            # Headers and status code only available for HTTP/S transfers
            self.status = response.statusCode()
            self.headers = dict(response.allHeaderFields())

        self.done = True

        if completionHandler:  # tell the session task to continue
            completionHandler(NSURLSessionResponseAllow)

    def URLSession_dataTask_didReceiveData_(
            self,
            session,  # type: NSURLSession
            task,     # type: NSURLSessionDataTask
            data      # type: NSData
    ):  # type: (...) -> None
        logger.debug('URLSession_dataTask_didReceiveData_ (%d bytes)', len(data))

        self.output.appendData_(data)
        self.bytesReceived += len(data)
        if self.expectedLength != NSURLResponseUnknownLength:
            self.percentComplete = int(
                float(self.bytesReceived)/float(self.expectedLength) * 100.0)

    # - NSURLSessionTaskDelegate

    def URLSession_task_didReceiveChallenge_completionHandler_(
            self,
            session,          # type: NSURLSession
            task,             # type: NSURLSessionTask
            challenge,        # type: NSURLAuthenticationChallenge
            completionHandler # type: (NSURLSessionAuthChallengeDisposition, NSURLCredential) -> Void
    ):  # type: (...) -> None
        logger.debug('URLSession_task_didReceiveChallenge_completionHandler_')
        completionHandler.__block_signature__ = objc_method_signature('v@i@')
        
        protectionSpace = challenge.protectionSpace()
        host = protectionSpace.host()
        realm = protectionSpace.realm()
        authenticationMethod = protectionSpace.authenticationMethod()

        logger.debug('NSURLProtectionSpace host: %s, realm: %s, method: %s', host,
                     realm, authenticationMethod)

        if authenticationMethod == 'NSURLAuthenticationMethodServerTrust' and not self.verify:
            logger.debug('Trusting invalid SSL certificate because verify=False')
            trust = protectionSpace.serverTrust()
            credential = NSURLCredential.credentialForTrust_(trust)
            completionHandler(
                NSURLSessionAuthChallengePerformDefaultHandling, credential)
        elif authenticationMethod in ['NSURLAuthenticationMethodDefault',
                                      'NSURLAuthenticationMethodHTTPBasic',
                                      'NSURLAuthenticationMethodHTTPDigest']:
            logger.debug('Attempting to authenticate')
            if getattr(self, 'credential', None) is not None:
                logger.debug('Using supplied NSURLCredential')
                completionHandler(
                    NSURLSessionAuthChallengeUseCredential, self.credential)
            else:
                logger.debug('No NSURLCredential available, not authenticating.')
                completionHandler(NSURLSessionAuthChallengePerformDefaultHandling, None)
        else:
            completionHandler(
                NSURLSessionAuthChallengePerformDefaultHandling, None)

    def URLSession_task_didCompleteWithError_(
            self,
            session,    # type: NSURLSession
            task,       # type: NSURLSessionTask
            error       # type: NSError
    ):  # type: (...) -> None
        logger.debug('URLSession_task_didCompleteWithError_')
        if error:
            self.error = error
            # If this was an SSL error, try to extract the SSL error code.
            if 'NSUnderlyingError' in error.userInfo():
                ssl_code = error.userInfo()['NSUnderlyingError'].userInfo().get(
                    '_kCFNetworkCFStreamSSLErrorOriginalValue', None)
                if ssl_code:
                    self.SSLError = (ssl_code, ssl_error_codes.get(
                        ssl_code, 'Unknown SSL error'))
        else:
            logger.debug('no error(s)')

        self.done = True

    def URLSession_task_willPerformHTTPRedirection_newRequest_completionHandler_(
            self,
            session,           # type: NSURLSession
            task,              # type: NSURLSessionTask
            response,          # type: NSHTTPURLResponse
            request,           # type: NSURLRequest
            completionHandler  # type: (NSURLRequest) -> None
    ):
        """This method copied largely from gurl.py"""
        # we don't actually use the session or task arguments, so
        # pylint: disable=W0613
        logger.debug('URLSession_task_willPerformHTTPRedirection_newRequest_')
        completionHandler.__block_signature__ = objc_method_signature('v@@')

        if response is None:
            # the request has changed the NSURLRequest in order to standardize
            # its format, for example, changing a request for
            # http://www.apple.com to http://www.apple.com/. This occurs because
            # the standardized, or canonical, version of the request is used for
            # cache management. Pass the request back as-is
            # (it appears that at some point Apple also defined a redirect like
            # http://developer.apple.com to https://developer.apple.com to be
            # 'merely' a change in the canonical URL.)
            # Further -- it appears that this delegate method isn't called at
            # all in this scenario, unlike NSConnectionDelegate method
            # connection:willSendRequest:redirectResponse:
            # we'll leave this here anyway in case we're wrong about that
            if completionHandler:
                completionHandler(request)
                return
            else:
                return request

        # If we get here, it appears to be a real redirect attempt
        # Annoyingly, we apparently can't get access to the headers from the
        # site that told us to redirect. All we know is that we were told
        # to redirect and where the new location is.
        newURL = request.URL().absoluteString()

        logging.debug('Being redirected to: %s' % newURL)
        # self.history.append()
        completionHandler(request)


class NSURLSessionAdapter(BaseAdapter):

    def __init__(self, credential=None, force_basic=False):
        # type: (Optional[NSURLCredential], bool) -> None
        """NSURLSessionAdapter implements a requests adapter using PyObjC + NSURLSession.

        Because of the way the delegate deals with authentication, we cannot support requests style auth which may only
        act on the content of the request. For NSURLSessionAdapter you should use NSURLCredentialAuth, which supplies an
        NSURLCredential object whenever authentication is required by the remote host.

        :param credential: NSURLCredential that will be used if the server requests basic auth.
        :param force_basic: Force basic authentication to be used for every request by preparing the Authorization
            header. This is sometimes required because NSURLSession only checks with the delegate for credentials when
            the server sends a "challenge". Some web services never even send a challenge.
        """
        super(NSURLSessionAdapter, self).__init__()

        self.verify = True
        self.credential = credential
        self.force_basic = force_basic
        self.configuration = None
        self.delegate = None
        self.session = None

    def send(
            self,
            request,       # type: PreparedRequest
            stream=False,  # type: bool
            timeout=None,  # type: Union[float, Tuple[float, float]]
            verify=True,   # type: bool
            cert=None,     # type: Any
            proxies=None   # type: dict
        ):
        # type: (...) -> Response

        self.configuration = NSURLSessionConfiguration.ephemeralSessionConfiguration()
        self.delegate = NSURLSessionAdapterDelegate.alloc().initWithAdapter_(self)
        self.delegate.credential = self.credential
        self.session = NSURLSession.sessionWithConfiguration_delegate_delegateQueue_(
            self.configuration,
            self.delegate,
            None,
        )

        nsrequest = build_request(request, timeout)
        self.verify = verify

        # TODO: Support all of this stuff.
        assert not stream
        assert not cert
        assert not proxies

        if request.method in ['PUT', 'POST'] and request.body is not None:
            # These verbs should usually be an upload task to send the correct request headers.
            task = self.session.uploadTaskWithRequest_fromData_(nsrequest, buffer(request.body))
        else:
            task = self.session.dataTaskWithRequest_(nsrequest)

        cookiestore = self.configuration.HTTPCookieStorage()

        task.resume()

        while not self.delegate.isDone():
            pass

        response = build_response(request, self.delegate, cookiestore)

        # if self.delegate.error is not None:
        #     raise self.delegate.error

        if self.delegate.SSLError is not None:
            code, message = self.delegate.SSLError
            raise SSLError('{}: {}'.format(code, message), response=response, request=request)

        return response



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""abc.py

This module describes abstract base classes used throughout python-jss.
Shea would call this non-pythonic but i'd like to verify exactly which subclasses are conforming 100% to
a declared interface - mosen.
"""

from __future__ import absolute_import
import abc


class AbstractRepository(object):

    # If these attributes arent supplied in the constructor, Repository raises a JSSError
    #required_attrs = set()

    # connection arguments
    # connection

    @abc.abstractmethod
    def copy_pkg(self, filename, id_=-1):  # type: (str, Optional[int]) -> None
        """Copy a package to the repo's Package subdirectory.

        Args:
            filename: Path for file to copy.
            id_: ID of Package object to associate with, or -1 for new
                packages (default). Only required for JDS
        """
        pass

    @abc.abstractmethod
    def delete(self, filename):  # type: (str) -> None
        """Delete a file from the repository.

        Args:
            filename: String filename only (i.e. no path) of file to
                delete.
        """
        pass

    @abc.abstractmethod
    def exists(self, filename):  # type: (str) -> bool
        """Report whether a file exists on the distribution point.

        Determines file type by extension.

        Args:
            filename: Filename you wish to check. (No path! e.g.:
                "AdobeFlashPlayer-14.0.0.176.pkg")
        """
        pass



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""python-jss

Python wrapper for the JAMF Casper JSS API.

"import jss" to import all public classes.

Public package contents include:
    casper: Class using the Casper private API call to casper.jxml.
    curl_adapter: Networking adapter to use the curl command with the
        Requests API.
    distribution_point: Classes for AFP, SMB, CDP, and JDS DPs.
    distribution_points: Class for managing distribution point classes.
        distribution points, JDS, and CDP distribution servers, and
        copying, deleting, and testing for files, as well as a
        controller class for abstracting all configured DPs.
    exceptions: python-jss custom exceptions.
    jamf_software_server: Class for representing a JSS, and for
        preference files to configure one.
    jssobject: Base class used for JSS objects. Useful for testing
        (e.g. "isinstance(obj, JSSObject)").
    jssobjects: Represents each of the objects the JSS supports
        (packages, computers, etc).
    jss_prefs: Class for loading python-jss configuration via a plist
        file, and for use as an argument to JSS. Includes an
        interactive setup helper.
    requests_adapter: Networking adapter to use the Requests library.
    response_adapter: Adapter for wrapping Curl response objects with the
        Requests API.

Private package contents include:
    contrib: Code from other authors used in python-jss.
    jssobjectlist: Classes for representing lists of objects returned
        from the JSS' GET searches.
    tlsadapter: Adapter to allow Requests to use TLS, and
        with the correct ciphers to match current JAMF recommendations.
    tools: Assorted functions for common tasks used throughout the
        package.
"""

from __future__ import absolute_import
from .casper import Casper
from .curl_adapter import CurlAdapter
from .distribution_point import (AFPDistributionPoint, SMBDistributionPoint,
                                 JDS, CDP, LocalRepository, JCDS, AWS)
from .distribution_points import DistributionPoints
from .exceptions import *
from .jamf_software_server import JSS
from .jssobject import JSSObject
from .jssobjects import *
from . import uapiobjects as uapi
from .jss_prefs import JSSPrefs
from .misc_endpoints import *
from .misc_uapi_endpoints import *
from .queryset import QuerySet
from .pretty_element import PrettyElement

import sys

sys.path.insert(0, '/Library/AutoPkg/JSSImporter')
import requests

from .tools import is_osx, is_linux, element_str

# Deprecated
from .jssobjectlist import JSSObjectList


__version__ = "2.1.0"



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""uapiobject.py

Base Classes representing JSS database objects and their UAPI endpoints
"""
from __future__ import print_function
from __future__ import absolute_import
from six import string_types

import os
import datetime as dt
import json

from jss import tools
try:
    from UserDict import UserDict  # Python 2.X
except ImportError:
    from collections import UserDict  # Python 3.3+

from .exceptions import JSSError, MethodNotAllowedError, PutError, PostError


DATE_FMT = "%Y/%m/%d-%H:%M:%S.%f"


class UAPIObject(UserDict):
    """Subclass for JSS UAPI objects which do not return a list of objects.

    These objects have in common that they cannot be created. They can,
    however, be updated.

    Class Attributes:
        cached: False, or datetime.datetime since last retrieval.
        can_get: Bool whether object allows a GET request.
        can_put: Bool whether object allows a PUT request.
        can_post: Bool whether object allows a POST request.
        can_delete: Bool whether object allows a DEL request.
    """
    _endpoint_path = None  # type: Optional[str]
    can_get = True         # type: bool
    can_put = True         # type: bool
    can_post = False       # type: bool
    can_delete = False     # type: bool

    def __init__(self, jss, data, **kwargs):
        # type: (JSS, Optional[dict], Optional[dict]) -> None
        """Initialize a new UAPIObject

        Args:
            jss (JSS, None): JSS to get data from, or None if no JSS
                communications need to be performed.
            data: dict data for the object
            kwargs: Unused, but present to support a unified signature
                for all subclasses (which need and use kwargs).
        """
        UserDict.__init__(self, data)
        self.jss = jss
        self.cached = False
        #super(UAPIObject, self).__init__(data)

    @property
    def cached(self):  # type: () -> bool
        # Check to see whether cache should be invalidated due to age.
        # If jss.max_age is 0, never cache.
        # If jss.max_age is negative, cache lasts forever.
        if isinstance(self._cached, dt.datetime) and self.jss.max_age > -1:
            max_age = dt.timedelta(seconds=self.jss.max_age)
            now = dt.datetime.now()
            if now - self._cached > max_age:
                self._cached = False

        return self._cached

    @cached.setter
    def cached(self, val):  # type: (bool) -> None
        self._cached = val

    @classmethod
    def build_query(cls, *args, **kwargs):
        # type: (Optional[List[any]], Optional[dict]) -> str
        """Return the path for query based on data type and contents.

        Args:
            *args, **kwargs: Left for compatibility with more
            full-featured subclasses' overrides.

        Returns:
            str path construction for this class to query.
        """
        return 'uapi/%s' % cls._endpoint_path

    @property
    def url(self):  # type: () -> str
        """Return the path subcomponent of the url to this object.

        For example: "/activationcode"
        """
        # Flat objects have no ID property, so there is only one URL.
        return 'uapi/%s' % self._endpoint_path

    def __repr__(self):
        if isinstance(self.cached, dt.datetime):
            cached = self.cached.strftime(DATE_FMT)
        else:
            cached = bool(self.cached)
        return "<{} cached: {} at 0x{:0x}>".format(
            self.__class__.__name__, cached, id(self))

    def retrieve(self):
        """Replace this object's data with JSS data, reset cache-age."""
        if self.jss.verbose:
            print("Retrieving data from JSS...")

        json_data = self.jss.get(self.url, headers={'Accept': 'application/json'})
        self.update(json_data)
        self.cached = dt.datetime.now()

    def save(self):
        """Update or create a new object on the JSS.

        If this object is not yet on the JSS, this method will create
        a new object with POST, otherwise, it will try to update the
        existing object with PUT.

        New: some UAPI endpoints require a POST even if they are a non CRUD object type. The presence of can_post = True
        will determine which method we use, with PUT always being the default.

        Data validation is up to the client; The JSS in most cases will
        at least give you some hints as to what is invalid.
        """
        if self.can_post:
            self.jss.post(self.url, data=self)
        else:
            self.jss.put(self.url, data=self)

        if self.can_get:
            # Replace current instance's data with new, JSS-validated data.
            self.retrieve()

    def to_file(self, path):
        """Write object JSON to path.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening.
        """
        with open(os.path.expanduser(path), "w") as ofile:
            ofile.write(json.dumps(self))

    def to_string(self):
        """Return indented object JSON as bytes."""
        return self.__str__()


class UAPIContainer(UAPIObject):
    """Subclass for members of a type of UAPI object.

    Class Attributes:
        cached: False, "Unsaved" for newly created objects that have
            not been POSTed to the JSS, or datetime.datetime since last
            retrieval.
        kwargs (dict): Keyword argument dictionary used in the original
            GET request to retrieve this object. By default, kwargs are
            applied to subsequent `retrieve()` operations unless
            cleared.
        can_get: Bool whether object allows a GET request.
        can_put: Bool whether object allows a PUT request.
        can_post: Bool whether object allows a POST request.
        can_delete: Bool whether object allows a DEL request.
        default_search: String default search type to utilize for GET.
        search_types: Dict of search types available to object:
            Key: Search type name. At least one must match the
                default_search.
            Val: URL component to use to request via this search_type.
        allowed_kwargs: Tuple of query extensions that are
            available (or sometimes required). Please see the JAMF Pro
            API documentation for the final word on how these work.
        data_keys: Dictionary of keys to create if instantiating a
            blank object using the _new method.
            Keys: String names of keys to create at top level.
            Vals: Values to set for the key.
                Int and bool values get converted to string.
                Dicts are recursively added (so their keys are added to
                    parent key, etc).

    Private Class Attributes:
        _id_path (str): URL Path subcomponent used to reference an
            object by ID when querying or posting.
    """
    can_get = True
    can_put = True
    can_post = True
    can_delete = True
    default_search = "name"  # type: str
    search_types = {"name": "name"}  # type: dict
    allowed_kwargs = tuple()  # type: Tuple[str]
    data_keys = {}  # type: dict
    _id_path = "id"  # type: str

    # Overrides ###############################################################
    def __init__(self, jss, data, **kwargs):
        """Initialize a new JSSObject from scratch or from a python dict.

        data still adheres to the Classic API convention of being able to pass a string which represents
        the object "name", as well as a dict, which will be the actual representation of the object.

        Args:
            jss (JSS, None): JSS object, or None if no communication
                with the JSS is needed.
            data (dict, Identity, str):
                data to use for creating the object, a name to use for
                creating a new object, or an Identity object representing
                basic object info.
            kwargs (str): Key/value pairs to be added to the object when
                building one from scratch.
        """
        self.jss = jss
        # self._basic_identity = Identity(name="", id="")
        self.kwargs = {}

        if isinstance(data, string_types):
            # self._new(data, **kwargs)
            UAPIObject.__init__(self, jss, {"name": data}, **kwargs)
            self.cached = "Unsaved"

        elif isinstance(data, dict):
            # Create a new object from passed XML.
            UAPIObject.__init__(self, jss, data, **kwargs)
            # If this has an ID, assume it's from the JSS and set the
            # cache time, otherwise set it to "Unsaved".
            if 'id' in data and data['id']:
                self.cached = dt.datetime.now()
            else:
                self.cached = "Unsaved"

        # elif isinstance(data, Identity):
        #     # This is basic identity information, probably from a
        #     # listing operation.
        #     new_xml = PrettyElement(tag=self.root_tag)
        #     super(Container, self).__init__(jss, new_xml)
        #     self._basic_identity = Identity(data)
        #     # Store any kwargs used in retrieving this object.
        #     self.kwargs = kwargs

        else:
            raise TypeError(
                "UAPIObjects data argument must be of type "
                "dict")

    def url(self, method='GET'):
        """Return the path subcomponent of the url to this object.

        For example: "computers/id/451"

        UAPI is not consistent with the endpoint name based on which operation you are performing.
        """
        if hasattr(self, '_endpoint_path_' + method.lower()):
            pass

        url_components = [self._endpoint_path, self._id_path, self.id]
        url_components.extend(self._process_kwargs(self.kwargs))
        return os.path.join(*url_components)

    @classmethod
    def _urlify_arg(cls, key, val):
        """Convert keyword arguments' values to proper format for GET"""
        if key == 'subset':
            if not isinstance(val, list):
                val = val.split("&")

            # If this is not a "basic" subset, and it's missing
            # "general", add it in, because we need the ID.
            if all(k not in val for k in ("general", "basic")):
                val.append("general")

            return ['subset', "&".join(val)]

        elif key == 'date_range':
            start, end = val
            fmt = lambda s: s.strftime('%Y-%m-%d')
            start = start if isinstance(start, string_types) else fmt(end)
            end = end if isinstance(end, string_types) else fmt(end)
            return ['{}_{}'.format(start, end)]

        else:
            return [key, val]
        
    @classmethod
    def _process_kwargs(cls, kwargs):
        kwarg_urls = []
        if kwargs and all(key in cls.allowed_kwargs for key in kwargs):
            kwargs = cls._handle_kwargs(kwargs)
            for key, val in kwargs.items():
                kwarg_urls.extend(cls._urlify_arg(key, val))
        return kwarg_urls

    @classmethod
    def _handle_kwargs(cls, kwargs):
        """Do nothing. Can be overriden by classes which need it."""
        return kwargs

    def _new(self, name, **kwargs):
        """Create a new UAPIObject with name and "keys".

        Generate a default Dict template for this object, based on
        the class attribute "keys".

        Args:
            name: String name of the object to use as the
                object's name property.
            kwargs:
                Accepted keyword args can be viewed by checking the
                "data_keys" class attribute. Typically, they include all
                top-level keys, and non-duplicated keys used elsewhere.

                Values will be cast to string. (Int 10, bool False
                become string values "10" and "false").

                Ignores kwargs that aren't in object's keys attribute.
        """
        super(UAPIContainer, self).__init__(self.jss, {"name": name})
        UserDict.__init__(self, kwargs)

        # # ignore _name_element
        # for k, v in self.data_keys.items():
        #     self[k] = v



from __future__ import absolute_import
from .jamf_software_server import JSS

UPLOAD_TYPE_MAC_PROFILE = 4
UPLOAD_TYPE_IOS_PROFILE = 22


class Upload(object):
    """This type of object represents all transfers carried out by the
    FrontEndUploadController (com.jamfsoftware.jss.fileupload.UploadController).

    These types of upload usually require the following:

    - A file identifier to indicate what type of file is being uploaded. This is a number corresponding to some kind
        of internal enum.
    - A session token, which is used as a CSRF token for each upload.
    - Request data is posted to /upload/?sessionIdentifier=&fileIdentifier=&session-token=
    - The response is an empty body (200).
    - The ajax call to monitorUploadProgress.ajax receives a redirect URL when the upload is finished.
    - The redirected URL must be saved to persist the upload.
    """
    def __init__(self, jss):  # type: (JSS) -> None
        self.jss = jss

    def monitor_progress(self):
        """The JAMF Pro Web UI often calls this endpoint to get upload status (to display a progress bar).
        The secondary functionality is that when the upload is finished, the payload may contain a redirect URI.
        For this reason we still need to use the endpoint to determine the upload outcome.

        The format of the response is an XML fragment, even though no Content-Type is specified.

        Example (for Push Cert .p12 upload):

            <jss>
                <javascript>nextPage</javascript>
                <status>complete</status>
                <filename>push.p12</filename>
                <sessionExpiresEpoch>1800</sessionExpiresEpoch>
            </jss>

        In this example, the `javascript` element includes a function to call.
        """
        self.jss.get('/monitorUploadProgress.ajax')

    def finalize(self):
        """POST to the legacy.html page representing the object, which confirms that the upload is complete.

        For this, it seems we need some of the hidden vars from the FORM previous:

        - session-token
        - INCLUDE_PAGE_VARIABLE (the server uses this to determine the previous step, and therefore what the current
            step requires).
        - OBJECT_RANDOM_IDENTIFIER (matches the identifier appended to the end of session_identifier)
        """
        pass


class UploadAssistant(object):
    """This class represents any assistant-style upload in the JAMF Pro UI.

    Although uploads are tracked by the `com.jamfsoftware.jss.fileupload.UploadController` there are sometimes other
    fields that must be submitted to make the database consistent.
    """
    def __init__(self, j, steps):  # type: (JSS, list) -> UploadAssistant
        self._jss = j
        self._step_index = 0
        self._steps = steps

    @property
    def step_index(self):
        return self._step_index

    @step_index.setter
    def step_index(self, v):
        self._step_index = v

    @property
    def step(self):
        return self._steps[self._step_index]


class PushNotificationKeystoreAssistant(UploadAssistant):

    FIELD_METHOD_DOWNLOAD_SIGNED_CSR = 1
    FIELD_METHOD_DOWNLOAD_CSR = 2
    FIELD_METHOD_UPLOAD_P12 = 3
    FIELD_METHOD_GET_PROXY_TOKEN = 4

    def __init__(self, j, **kwargs):
        super(UploadAssistant, self).__init__(j, [
            'pushCertificateChooseMethod.jsp',
            'pushCertificateJAMFNationLogin.jsp',
            'pushCertificateDownloadSignedCSR.jsp',
            'pushCertificateDownloadUnsignedCSR.jsp',
            'pushCertificateGenerateCSRManually.jsp',
            'pushCertificateCreateCertificate.jsp',
            'pushCertificateUploadCertificate.jsp',
            'pushCertificateUploadP12.jsp',
            'pushCertificateEnterPasswordForP12.jsp',
        ], **kwargs)

    def download_signed_csr(self, jamf_nation_username, jamf_nation_password):
        pass

    def download_csr(self):
        pass

    def upload_p12(self, path, passphrase=None):
        pass

    def get_proxy_token(self):
        pass


class MacOSConfigurationProfileUpload(Upload):
    """This object represents a direct upload of a macOS configuration profile to the web ui.

    Use this to upload signed profiles directly into the JPS.
    """

    def __init__(self, jss):
        super(MacOSConfigurationProfileUpload, self).__init__(jss)
        self.session_identifier = "UPLOAD_FROM_LIST"
        self.file_identifier = UPLOAD_TYPE_MAC_PROFILE


class IOSConfigurationProfileUpload(Upload):

    def __init__(self, jss):
        super(IOSConfigurationProfileUpload, self).__init__(jss)
        self.session_identifier = "UPLOAD_FROM_LIST"
        self.file_identifier = UPLOAD_TYPE_IOS_PROFILE


class PushNotificationKeystoreUpload(Upload):

    def __init__(self, jss):
        """Construct a PushNotificationKeystoreUpload (Upload a .p12 file as push keystore).

        Notes:

        - The sessionIdentifier is constructed like so (see WEB-INF/frontend/standardFormUpload.jsp):

                sessionIdentifier = target.getClass().getName() + ":" + target.getRandomIdentifier();

            In this case the classname becomes `com.jamfsoftware.jss.objects.pushnotification.PushNotificationKeystore`.

        - The fileIdentifier is always "Keystore" (It is hardcoded into the .jsp,
            see WEB-INF/frontend/pushCertificateUploadP12.jsp).

        - The object random identifier is generated at page load time, so we have to scrape it at least once.

        Sequence:

            - Hit `legacy/pushNotificationCertificate.html?id=-1&o=c`, which is usually in an IFRAME.
            - Scrape hidden inputs from <form name="f" id="f"> :
                - name="session-token" id="session-token"
                - name="INCLUDE_PAGE_VARIABLE" id="INCLUDE_PAGE_VARIABLE"
                - name="OBJECT_RANDOM_IDENTIFIER" id="OBJECT_RANDOM_IDENTIFIER"
            - POST all of above, with FIELD_METHOD: 3 (upload p12) and action=Next
            - POST multipart/form-data file as `importFileKeystore` to:
                - upload/?sessionIdentifier=com.jamfsoftware.jss.objects.pushnotification.PushNotificationKeystore:<randomid>&fileIdentifier=Keystore"
            - When the monitor ajax endpoint says complete, POST again to legacy/pushNotificationCertificate.html with
                the INCLUDE_PAGE_VARIABLE from the previous step.
            - Now we are entering the keystore password
            - POST the keystore password as `FIELD_UPLOADED_P12_PASSWORD` and include the INCLUDE_PAGE_VARIABLE from the
                last page. with action=Next
            - Done.

        """
        super(PushNotificationKeystoreUpload, self).__init__(jss)
        self.session_identifier = "com.jamfsoftware.jss.objects.pushnotification.PushNotificationKeystore"
        self.file_identifier = "Keystore"
        self.object_random_identifier = ""

    def finalize_p12_passphrase(self):
        """The upload of a .p12 keystore has a second phase of finalization which is to specify the passphrase.

        The passphrase is included in a form variable `FIELD_UPLOADED_P12_PASSWORD`, as well as the variables from
        the previous POST to upload:

        - session-token
        - INCLUDE_PAGE_VARIABLE
        - OBJECT_RANDOM_IDENTIFIER
        - fakeUsername (empty)
        - fakePassword (empty)
        - action = Next
        """
        pass



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""uapi.py

Classes representing UAPI endpoints that are more RPC style than CRUD.
"""


__all__ = 'SystemInitialize',


class SystemInitialize(object):
    _endpoint_path = "system/initialize"
    can_get = False
    can_put = False
    can_delete = False

    def __init__(self, jss):
        """Initialize a new SystemInitialize

        Args:
            jss: JSS object.
        """
        self.jss = jss

    @property
    def url(self):
        """Return the path subcomponent of the url to this object."""
        return self._endpoint_path

    def initialize(self, data):
        r = self.jss.post(self.url, data)


class RecalculateComputerSmartGroups(object):
    _endpoint_path = "computer/obj/computer"
    can_get = False
    can_put = False
    can_delete = False


class RecalculateSmartGroupComputers(object):
    _endpoint_path = "computer/obj/smartgroup"
    can_get = False
    can_put = False
    can_delete = False



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""misc_endpoints.py

Classes representing API endpoints that don't subclass JSSObject
"""
from __future__ import print_function

from __future__ import absolute_import
import mimetypes
import os
import sys
from xml.etree import ElementTree

from .exceptions import MethodNotAllowedError, PostError
from .tools import error_handler


__all__ = ('CommandFlush', 'FileUpload', 'LogFlush')

# Map Python 2 basestring type for Python 3.
if sys.version_info.major == 3:
    basestring = str

# pylint: disable=missing-docstring
# pylint: disable=too-few-public-methods

class CommandFlush(object):
    _endpoint_path = "commandflush"
    can_get = False
    can_put = False
    can_post = False

    def __init__(self, jss):
        """Initialize a new CommandFlush

        Args:
            jss: JSS object.
        """
        self.jss = jss

    @property
    def url(self):
        """Return the path subcomponent of the url to this object."""
        return self._endpoint_path

    def command_flush_with_xml(self, data):
        """Flush commands for devices with a supplied xml string.

        From the Casper API docs:
        Status and devices specified in an XML file. Id lists may be
        specified for <computers>, <computer_groups>, <mobile_devices>,
        <mobile_device_groups>. Sample file:
            <commandflush>
              <status>Pending+Failed</status>
              <mobile_devices>
                <mobile_device>
                  <id>1</id>
                </mobile_device>
                <mobile_device>
                  <id>2</id>
                </mobile_device>
              </mobile_devices>
            </commandflush>

        Args:
            data (string): XML string following the above structure or
                an ElementTree/Element.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        if not isinstance(data, basestring):
            data = ElementTree.tostring(data, encoding='UTF-8')
        self.jss.delete(self.url, data)

    def command_flush_for(self, id_type, command_id, status):
        """Flush commands for an individual device.

        Args:
            id_type (str): One of 'computers', 'computergroups',
                'mobiledevices', or 'mobiledevicegroups'.
            id_value (str, int, list): ID value(s) for the devices to
                flush. More than one device should be passed as IDs
                in a list or tuple.
            status (str): One of 'Pending', 'Failed', 'Pending+Failed'.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        id_types = ('computers', 'computergroups', 'mobiledevices',
                    'mobiledevicegroups')
        status_types = ('Pending', 'Failed', 'Pending+Failed')
        if id_type not in id_types or status not in status_types:
            raise ValueError("Invalid arguments.")

        if isinstance(command_id, list):
            command_id = ",".join(str(item) for item in command_id)

        flush_url = "{}/{}/id/{}/status/{}".format(
            self.url, id_type, command_id, status)

        self.jss.delete(flush_url)

# pylint: disable=too-few-public-methods


class FileUpload(object):
    """FileUploads are a special case in the API. They allow you to add
    file resources to a number of objects on the JSS.

    To use, instantiate a new FileUpload object, then use the save()
    method to upload.

    Once the upload has been posted you may only interact with it
    through the web interface. You cannot list/get it or delete it
    through the API.

    However, you can reuse the FileUpload object if you wish, by
    changing the parameters, and issuing another save().
    """
    _endpoint_path = "fileuploads"
    allowed_kwargs = ('subset',)

    def __init__(self, j, resource_type, id_type, _id, resource):
        """Prepare a new FileUpload.

        Args:
            j: A JSS object to POST the upload to.
            resource_type:
                String. Acceptable Values:
                    Attachments:
                        computers
                        mobiledevices
                        enrollmentprofiles
                        peripherals
                        mobiledeviceenrollmentprofiles
                    Icons:
                        policies
                        ebooks
                        mobiledeviceapplicationsicon
                    Mobile Device Application:
                        mobiledeviceapplicationsipa
                    Disk Encryption
                        diskencryptionconfigurations
                        diskencryptions (synonymous)
                    PPD
                        printers
            id_type:
                String of desired ID type:
                    id
                    name
            _id: Int or String referencing the identity value of the
                resource to add the FileUpload to.
            resource: String path to the file to upload.
        """
        resource_types = ["computers", "mobiledevices", "enrollmentprofiles",
                          "peripherals", "mobiledeviceenrollmentprofiles",
                          "policies", "ebooks", "mobiledeviceapplicationsicon",
                          "mobiledeviceapplicationsipa",
                          "diskencryptionconfigurations", "printers"]
        id_types = ["id", "name"]

        self.jss = j

        # Do some basic error checking on parameters.
        if resource_type in resource_types:
            self.resource_type = resource_type
        else:
            raise TypeError(
                "resource_type must be one of: %s" % ', '.join(resource_types))
        if id_type in id_types:
            self.id_type = id_type
        else:
            raise TypeError("id_type must be one of: %s" % ', '.join(id_types))
        self._id = str(_id)

        basename = os.path.basename(resource)
        content_type = mimetypes.guess_type(basename)[0]
        self.resource = {"name": (basename, open(resource, "rb"),
                                  content_type)}
        self._set_upload_url()

    def _set_upload_url(self):
        """Generate the full URL for a POST."""
        # pylint: disable=protected-access
        self._upload_url = "/".join([
            self.jss._url, self._endpoint_path, self.resource_type,
            self.id_type, str(self._id)])
        # pylint: enable=protected-access

    def save(self):
        """POST the object to the JSS."""
        try:
            response = self.jss.session.post(
                self._upload_url, files=self.resource)
        except PostError as error:
            if error.status_code == 409:
                raise PostError(error)
            else:
                raise MethodNotAllowedError(self.__class__.__name__)

        if response.status_code == 201:
            if self.jss.verbose:
                print("POST: Success")
                print(response.content)
        elif response.status_code >= 400:
            error_handler(PostError, response)


class LogFlush(object):
    _endpoint_path = "logflush"

    def __init__(self, jss):
        """Initialize a new LogFlush

        Args:
            jss: JSS object.
        """
        self.jss = jss

    @property
    def url(self):
        """Return the path subcomponent of the url to this object."""
        return self._endpoint_path

    def log_flush_with_xml(self, data):
        """Flush logs for devices with a supplied xml string.

        From the Casper API docs:
            log, log_id, interval, and devices specified in an XML file.
            Sample file:
              <logflush>
                <log>policy</log>
                <log_id>2</log_id>
                <interval>THREE MONTHS</interval>
                <computers>
                  <computer>
                    <id>1</id>
                  </computer>
                  <computer>
                    <id>2</id>
                  </computer>
                </computers>
              </logflush>

        Args:
            data (string): XML string following the above structure or
                an ElementTree/Element.
                Elements:
                    logflush (root)
                    log (Unknown; "policy" is the only one listed in
                         docs).
                    log_id: Log ID value.
                     interval: Combination of "Zero", "One", "Two",
                        "Three", "Six", and "Day", "Week", "Month",
                        "Year". e.g. ("Three+Months")
                        Please note: The documentation for this
                        specifies the singular form (e.g. "Month"),
                        and plural ("Months") at different times, and
                        further the construction is listed as
                        "THREE MONTHS" elsewhere. Limited testing
                        indicates that pluralization does not matter,
                        nor does capitalization. The "+" seems optional
                        as well.
                        Please test!
                    Device Arrays:
                        Again, acceptable values are not listed in the
                        docs, aside from the example ("computers").
                        Presumably "mobiledevices", and possibly
                        "computergroups" and "mobiledevicegroups" work.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        if not isinstance(data, basestring):
            data = ElementTree.tostring(data, encoding='UTF-8')
        self.jss.delete(self.url, data)

    def log_flush_for_interval(self, log_type, interval):
        """Flush logs for an interval of time.

        Args:
            log_type (str): Only documented type is "policies". This
                will be applied by default if nothing is passed.
            interval (str): Combination of "Zero", "One", "Two",
                "Three", "Six", and "Day", "Week", "Month", "Year". e.g.
                ("Three+Months") Please note: The documentation for this
                specifies the singular form (e.g. "Month"), and plural
                ("Months") at different times, and further the
                construction is listed as "THREE MONTHS" elsewhere.
                Limited testing indicates that pluralization does not
                matter, nor does capitalization.
                Please test!

                No validation is performed on this prior to the request
                being made.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        if not log_type:
            log_type = "policies"

        # The XML for the /logflush basic endpoint allows spaces
        # instead of "+", so do a replace here just in case.
        interval = interval.replace(" ", "+")

        flush_url = "{}/{}/interval/{}".format(
            self.url, log_type, interval)

        self.jss.delete(flush_url)

    def log_flush_for_obj_for_interval(self, log_type, obj_id, interval):
        """Flush logs for an interval of time for a specific object.

        Please note, log_type is a variable according to the API docs,
        but acceptable values are not listed. Only "policies" is
        demonstrated as an acceptable value.

        Args:
            log_type (str): Only documented type is "policies". This
                will be applied by default if nothing is passed.
            obj_id (str or int): ID of the object to have logs flushed.
            interval (str): Combination of "Zero", "One", "Two",
                "Three", "Six", and "Day", "Week", "Month", "Year". e.g.
                ("Three+Months") Please note: The documentation for this
                specifies the singular form (e.g. "Month"), and plural
                ("Months") at different times, and further the
                construction is listed as "THREE MONTHS" elsewhere.
                Limited testing indicates that pluralization does not
                matter, nor does capitalization.
                Please test!

                No validation is performed on this prior to the request
                being made.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        if not log_type:
            log_type = "policies"

        # The XML for the /logflush basic endpoint allows spaces
        # instead of "+", so do a replace here just in case.
        interval = interval.replace(" ", "+")

        flush_url = "{}/{}/id/{}/interval/{}".format(
            self.url, log_type, obj_id, interval)

        self.jss.delete(flush_url)


# pylint: enable=missing-docstring
# pylint: enable=too-few-public-methods



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""queryset.py

Class that adds some extra functionality to a basic list. Used as the
result of all queries in python-jss.
"""


from __future__ import absolute_import
from collections import defaultdict
try:
    import cPickle  # Python 2.X
except ImportError:
    import _pickle as cPickle  # Python 3+
import datetime
import os

from .jssobject import DATE_FMT, Identity


STR_FMT = "{0:>{1}} | {2:>{3}} | {4:>{5}}"


class QuerySet(list):
    """A list style collection of JSSObjects.

    Listing operations retrieve minimal or overview information for most
    object types. For example, we may want to see all the Computers on
    the JSS but that does not mean we want to do a full object GET for
    each one.

    QuerySets hold instances of a single type of JSSObject, and use the
    python list API, while adding some extra helper-methods on top.
    """

    def __init__(self, objects):
        """Construct a list of JSSObjects.

        Args:
            objects (sequence of JSSObjects):
                Sequences must be of a single class.
        """
        if objects and not len({i.__class__ for i in objects}) == 1:
            raise ValueError
        super(QuerySet, self).__init__(objects)
        self.sort()
        self.contained_class = objects[0].__class__ if objects else None

    def __str__(self):
        """Make data human readable."""
        # Make ID, Name first, no matter what.
        sort_keys = ["id", "name"]
        if self:
            sort_keys.extend([
                key for key in self[0]._basic_identity.keys() if
                key not in sort_keys])

        # Build a dict of max lengths per column for table output.
        lengths = defaultdict(int)
        for item in self:
            for key in sort_keys:
                val = item._basic_identity[key] or ""
                length = max(len(key), len(val))
                if length > lengths[key]:
                    lengths[key] = length

        # Build a format string for row output.
        format_strings = []
        for key in sort_keys:
            length = lengths[key]
            format_strings.append("{{data[{}]:>{}}}".format(key, length))

        cached = 'cached'
        cached_format = '| {{cached:>{}}} |'.format(len(cached))

        fmt = "| " + " | ".join(format_strings) + cached_format

        # Begin building output with header lines.
        # Contained class can be None
        contained_name = self.contained_class.__name__ if self.contained_class is not None else "Empty"
        results = ["{} QuerySet".format(contained_name)]
        headers = {key: key for key in lengths}
        header_line = fmt.format(data=headers, cached="cached")
        bar = len(header_line) * '-'
        results.extend([bar, header_line, bar])

        str_cached = (
            lambda i: str(i.cached) if isinstance(i.cached, bool) else 'True')

        table = [
            fmt.format(data=item._basic_identity, cached=str_cached(item)) for
            item in self]
        results.extend(table)

        results.append(bar)
        return "\n".join(results)

    def __repr__(self):
        """Make data human readable."""
        return "QuerySet({})".format(super(QuerySet, self).__repr__())

    def sort(self, **kwargs):
        """Sort list elements by ID.
        """
        super(QuerySet, self).sort(key=lambda k: int(k.id))

    def sort_by_name(self):
        """Sort list elements by name."""
        super(QuerySet, self).sort(key=lambda k: k.name.upper())

    def retrieve_all(self):
        """Tell each contained object to retrieve its data from the JSS

        This can take a long time given a large number of objects,
        and depending on the size of each object.

        Returns:
            self (QuerySet) to allow method chaining.
        """
        for obj in self:
            if not obj.cached:
                obj.retrieve()

        return self

    def save_all(self):
        """Tell each contained object to save its data to the JSS

        This can take a long time given a large number of objects,
        and depending on the size of each object.

        Returns:
            self (QuerySet) to allow method chaining.
        """
        for obj in self:
            obj.save()

        return self

    def invalidate(self):
        """Clear the cache datetime for all contents.

        This causes objects to retrieve their data again when accessed.
        """
        for i in self: i.cached = False

    def names(self):
        """Return a generator of contents names"""
        return (item.name for item in self)

    def ids(self):
        """Return a generator of contents ids"""
        return (item.id for item in self)

    @classmethod
    def from_response(cls, obj_class, response, jss=None, **kwargs):
        """Build a QuerySet from a listing Response."""
        response_objects = (
            i for i in response if i is not None and i.tag != "size")

        dicts = (
            {child.tag: child.text for child in item} for item in
            response_objects)
        identities = (Identity(d) for d in dicts)

        objects = [obj_class(jss, data=i, **kwargs) for i in identities]

        return cls(objects)



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""casper.py

Utility class for getting and presenting information from casper.jxml.

The results from casper.jxml are undocumented and thus quite likely to be
removed. Do not rely on its continued existence!
"""
from __future__ import unicode_literals

from __future__ import absolute_import
try:
    # Python 2
    from __builtin__ import str as text
except ImportError:
    # Python 3
    from builtins import str as text

# 2 and 3 compatible
try:
    from urllib.parse import urlparse, urlencode
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError

from xml.etree import ElementTree

from .pretty_element import PrettyElement


class Casper(ElementTree.Element):
    """Interact with the JSS through its private casper endpoint.

    The API user must have the Casper Admin privileges "Use Casper
    Admin" and "Save With Casper Admin".
    """

    def __init__(self, jss):
        """Initialize a Casper object.

        Args:
            jss: A JSS object to request the casper page from.
        """
        self.jss = jss
        self.url = "%s/casper.jxml" % self.jss.base_url

        # This may be incorrect, but the assumption here is that this
        # request wants the auth information as urlencoded data; and
        # urlencode needs bytes.
        # TODO: If we can just pass in bytes rather than
        # urlencoded-bytes, we can remove this and let the request
        # adapter handle the outgoing data encoding.
        user = text(self.jss.user)
        password = text(self.jss.password)
        self.auth = urlencode(
            {"username": user, "password": password})
        super(Casper, self).__init__("Casper")
        self.update()

    def update(self):
        """Request an updated set of data from casper.jxml."""
        response = self.jss.session.post(
            self.url, data=self.auth)
        response_xml = ElementTree.fromstring(response.content)

        # Remove previous data, if any, and then add in response's XML.
        self.clear()
        self.extend(response_xml)




# encoding: utf-8
#
# Copyright 2009-2020 Greg Neagle.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FoundationPlist.py -- a tool to generate and parse OS X .plist files.

This is intended as a drop-in replacement for Python's included plistlib,
with a few caveats:
    - readPlist() and writePlist() operate only on a filepath,
        not a file object.
    - there is no support for the deprecated functions:
        readPlistFromResource()
        writePlistToResource()
    - there is no support for the deprecated Plist class.

The Property List (.plist) file format is a simple XML pickle supporting
basic object types, like dictionaries, lists, numbers and strings.
Usually the top level object is a dictionary.

To write out a plist file, use the writePlist(rootObject, filepath)
function. 'rootObject' is the top level object, 'filepath' is a
filename.

To parse a plist from a file, use the readPlist(filepath) function,
with a file name. It returns the top level object (again, usually a
dictionary).

To work with plist data in strings, you can use readPlistFromString()
and writePlistToString().
"""
from __future__ import absolute_import, print_function

# PyLint cannot properly find names inside Cocoa libraries, so issues bogus
# No name 'Foo' in module 'Bar' warnings. Disable them.
# pylint: disable=E0611
from Foundation import NSData
from Foundation import NSPropertyListSerialization
from Foundation import NSPropertyListMutableContainers
from Foundation import NSPropertyListXMLFormat_v1_0
# pylint: enable=E0611

# Disable PyLint complaining about 'invalid' camelCase names
# pylint: disable=C0103


class FoundationPlistException(Exception):
    """Basic exception for plist errors"""
    pass

class NSPropertyListSerializationException(FoundationPlistException):
    """Read/parse error for plists"""
    pass

class NSPropertyListWriteException(FoundationPlistException):
    """Write error for plists"""
    pass

def readPlist(filepath):
    """
    Read a .plist file from filepath.  Return the unpacked root object
    (which is usually a dictionary).
    """
    plistData = NSData.dataWithContentsOfFile_(filepath)
    dataObject, dummy_plistFormat, error = (
        NSPropertyListSerialization.
        propertyListFromData_mutabilityOption_format_errorDescription_(
            plistData, NSPropertyListMutableContainers, None, None))
    if dataObject is None:
        if error:
            error = error.encode('ascii', 'ignore')
        else:
            error = "Unknown error"
        errmsg = "%s in file %s" % (error, filepath)
        raise NSPropertyListSerializationException(errmsg)
    else:
        return dataObject


def readPlistFromString(data):
    '''Read a plist data from a (byte)string. Return the root object.'''
    plistData = NSData.dataWithBytes_length_(data, len(data))
    if not plistData:
        raise NSPropertyListSerializationException(
            "Could not convert string to NSData")
    dataObject, dummy_plistFormat, error = (
        NSPropertyListSerialization.
        propertyListFromData_mutabilityOption_format_errorDescription_(
            plistData, NSPropertyListMutableContainers, None, None))
    if dataObject is None:
        if error:
            error = error.encode('ascii', 'ignore')
        else:
            error = "Unknown error"
        raise NSPropertyListSerializationException(error)
    else:
        return dataObject


def writePlist(dataObject, filepath):
    '''
    Write 'rootObject' as a plist to filepath.
    '''
    plistData, error = (
        NSPropertyListSerialization.
        dataFromPropertyList_format_errorDescription_(
            dataObject, NSPropertyListXMLFormat_v1_0, None))
    if plistData is None:
        if error:
            error = error.encode('ascii', 'ignore')
        else:
            error = "Unknown error"
        raise NSPropertyListSerializationException(error)
    else:
        if plistData.writeToFile_atomically_(filepath, True):
            return
        else:
            raise NSPropertyListWriteException(
                "Failed to write plist data to %s" % filepath)


def writePlistToString(rootObject):
    '''Return 'rootObject' as a plist-formatted (byte)string.'''
    plistData, error = (
        NSPropertyListSerialization.
        dataFromPropertyList_format_errorDescription_(
            rootObject, NSPropertyListXMLFormat_v1_0, None))
    if plistData is None:
        if error:
            error = error.encode('ascii', 'ignore')
        else:
            error = "Unknown error"
        raise NSPropertyListSerializationException(error)
    else:
        return bytes(plistData)


if __name__ == '__main__':
    print('This is a library of support tools for the Munki Suite.')



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jssobjectlist.py

Deprecated in favor of QuerySet.
"""


from __future__ import absolute_import
import warnings

from .queryset import QuerySet


class JSSObjectList(QuerySet):
    """Deprecated in favor of QuerySet."""

    def __init__(self, factory, obj_class, objects=None):
        super(JSSObjectList, self).__init__(objects)
        warnings.warn(
            'JSSObjectList is deprecated and will be removed in the future. '
            'Please update code to use jss.QuerySet.', FutureWarning,
            stacklevel=2)



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jssobject.py

Base Classes representing JSS database objects and their API endpoints
"""
from __future__ import print_function
from __future__ import absolute_import
from six import string_types

import collections
import copy
import sys

try:
    import cPickle  # Python 2.X
except ImportError:
    import _pickle as cPickle  # Python 3+

import datetime as dt
import gzip
import os
from xml.etree import ElementTree

from .exceptions import JSSError, MethodNotAllowedError, PutError, PostError
from .pretty_element import PrettyElement
from jss import tools



DATE_FMT = "%Y/%m/%d-%H:%M:%S.%f"
_MATCH = "match"

# Map Python 2 unicode type for Python 3.
if sys.version_info.major == 3:
    unicode = str

class Identity(dict):
    """Subclass of dict used simply for type-checking."""
    pass


class JSSObject(PrettyElement):
    """Subclass for JSS objects which do not return a list of objects.

    These objects have in common that they cannot be created. They can,
    however, be updated.

    Attributes:
        cached (:obj:`datetime.datetime`, optional): False, or datetime.datetime since last retrieval.
        can_get (bool): whether object allows a GET request.
        can_put (bool): whether object allows a PUT request.
        can_post (bool): whether object allows a POST request.
        can_delete (bool): whether object allows a DEL request.

    Args:
        jss (:obj:`JSS`, optional): JSS to get data from, or None if no JSS
            communications need to be performed.
        data (:obj:`xml.etree.ElementTree.Element`): data for the object
        **kwargs: Unused, but present to support a unified signature
            for all subclasses (which need and use kwargs).
    """
    _endpoint_path = None
    can_get = True
    can_put = True
    can_post = False
    can_delete = False

    __str__ = tools.triggers_cache(tools.element_str)

    def __init__(self, jss, data, **kwargs):
        self.jss = jss
        self.cached = False

        # Turn Elements into PrettyElements (adds pretty printing
        # and fancy attribute finding).
        super(JSSObject, self).__init__(data)

    @property
    def cached(self):
        # Check to see whether cache should be invalidated due to age.
        # If jss.max_age is 0, never cache.
        # If jss.max_age is negative, cache lasts forever.
        if isinstance(self._cached, dt.datetime) and self.jss.max_age > -1:
            max_age = dt.timedelta(seconds=self.jss.max_age)
            now = dt.datetime.now()
            if now - self._cached > max_age:
                self._cached = False

        return self._cached

    @cached.setter
    def cached(self, val):
        self._cached = val

    @classmethod
    def build_query(cls, *args, **kwargs):
        """Return the path for query based on data type and contents.

        Args:
            *args, **kwargs: Left for compatibility with more
            full-featured subclasses' overrides.

        Returns:
            str path construction for this class to query.
        """
        return 'JSSResource/%s' % cls._endpoint_path

    @property
    def url(self):
        """Return the path subcomponent of the url to this object.

        For example: "/activationcode"
        """
        # Flat objects have no ID property, so there is only one URL.
        return 'JSSResource/%s' % self._endpoint_path

    def __repr__(self):
        if isinstance(self.cached, dt.datetime):
            cached = self.cached.strftime(DATE_FMT)
        else:
            cached = bool(self.cached)
        return "<{} cached: {} at 0x{:0x}>".format(
            self.__class__.__name__, cached, id(self))

    def __eq__(self, other):
        # There is no way to really compare as equal without grabbing
        # full data, so trigger a retrieval with `str()`
        return other.__class__ == self.__class__ and str(self) == str(other)

    def __hash__(self):
        # There is no way to really compare as equal without grabbing
        # full data, so trigger a retrieval with `str()`
        return hash(str(self))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Try to save on the way out of the with statement."""
        self.save()

    def _reset_data(self, updated_data):
        """Clear all children of base element and replace with update"""
        self.clear()
        # Convert all incoming data to PrettyElements.
        for child in updated_data.getchildren():
            if not isinstance(child, PrettyElement):
                child = PrettyElement(child)
            self._children.append(child)

    def retrieve(self):
        """Replace this object's data with JSS data, reset cache-age."""
        if self.jss.verbose:
            print("Retrieving data from JSS...")

        xmldata = self.jss.get(self.url)
        self._reset_data(xmldata)
        self.cached = dt.datetime.now()

    def save(self):
        """Update or create a new object on the JSS.

        If this object is not yet on the JSS, this method will create
        a new object with POST, otherwise, it will try to update the
        existing object with PUT.

        Data validation is up to the client; The JSS in most cases will
        at least give you some hints as to what is invalid.
        """
        try:
            self.jss.put(self.url, data=self)
        except PutError as put_error:
            # Something when wrong.
            raise PutError(put_error)

        # Replace current instance's data with new, JSS-validated data.
        self.retrieve()

    def to_file(self, path):
        """Write object XML to path.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening.
        """
        with open(os.path.expanduser(path), "w") as ofile:
            ofile.write(str(self))

    def to_string(self):
        """Return indented object XML as bytes."""
        return self.__str__()

    def pickle(self, path):
        """Write object to python pickle.

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening.
        """
        with open(os.path.expanduser(path), "wb") as pickle:
            cPickle.Pickler(pickle, cPickle.HIGHEST_PROTOCOL).dump(self)

    def tree(self, depth=None):
        """Return a formatted string representing object's tags

        This method removes redundent entries, so you only see
        unique keys.

        Args:
            depth (int): Depth of the tree to represent. Defaults to
                showing the entire tree.

        Returns:
            str with proper indention and newlines embedded, ready for
            printing.
        """
        results = self._get_tags(self, depth)
        return '\n'.join(results)

    def _get_tags(self, element, depth, level=0):
        results = []
        space = ' '
        indent_size = 4
        if depth is None or level < depth:
            for child in element:
                entry = space * indent_size * level + child.tag
                if entry not in results:
                    results.append(entry)
                    if len(child):
                        results.extend(self._get_tags(child, depth, level + 1))

        return results


# Decorate all public API methods that should trigger a retrieval of the
# object's full data from the JSS.
cache_triggers = (
    '__getitem__', '__len__', '__setitem__', '__str__', 'copy',
    'extend', 'find', 'findall', 'findtext', 'get', 'getchildren',
    'getiterator', 'insert', 'items', 'iter', 'iterfind', 'itertext', 'keys',
    'remove', 'set')

# Ones that block us:
# - Are not methods: 'tail', 'text', 'attrib','tag'
# - Used in setup: 'append',,'__setattr__', 'append',
tools.decorate_class_with_caching(JSSObject, cache_triggers)


class Container(JSSObject):
    """Subclass for members of a type of object.

    This includes the majority of objects in the JSS, for example
    Computers and Policies.

    Attributes:
        cached (:obj:`datetime.datetime`, optional): False, or datetime.datetime since last retrieval.
        can_get (bool): whether object allows a GET request.
        can_put (bool): whether object allows a PUT request.
        can_post (bool): whether object allows a POST request.
        can_delete (bool): whether object allows a DEL request.
        **kwargs: Keyword argument dictionary used in the original
            GET request to retrieve this object. By default, kwargs are
            applied to subsequent `retrieve()` operations unless
            cleared.
        default_search (str): String default search type to utilize for GET.
        search_types (dict): Dict of search types available to object:
            Key: Search type name. At least one must match the
                default_search.
            Val: URL component to use to request via this search_type.
        allowed_kwargs: Tuple of query extensions that are
            available (or sometimes required). Please see the Casper
            API documentation for the final word on how these work.
        root_tag (str): String singular form of object type found in
            containers (e.g. ComputerGroup has a container with tag:
            "computers" holding "computer" elements. The root_tag is
            "computer").
        data_keys (dict): Dictionary of keys to create if instantiating a
            blank object using the _new method.
            Keys: String names of keys to create at top level.
            Vals: Values to set for the key.
                Int and bool values get converted to string.
                Dicts are recursively added (so their keys are added to
                    parent key, etc).
        _id_path (str): URL Path subcomponent used to reference an
            object by ID when querying or posting.
        _name_element (str): XML path to element which contains
            the name of the object, used for creating new objects only.

            Most objects use a name tag at the root, (e.g. "name")
            Some, however, put it into "general/name". If you are
            implementing `data_keys` for an object type not yet
            impelemented, make sure to set this if it differs from the
            default/inherited value.

    Args:
        jss (:obj:`JSS`, optional): JSS to get data from, or None if no JSS
            communications need to be performed.
        data (:obj:`xml.etree.ElementTree.Element`, :obj:`Identity`, str): XML
            data to use for creating the object, a name to use for
            creating a new object, or an Identity object representing
            basic object info.
        **kwargs: Key/value pairs to be added to the object when
            building one from scratch.
    """
    root_tag = "Container"
    can_get = True
    can_put = True
    can_post = True
    can_delete = True
    default_search = "name"
    search_types = {"name": "name"}
    allowed_kwargs = tuple()
    data_keys = {}
    _id_path = "id"
    # TODO: Determine correct values for all endpoints.
    _name_element = "name"

    # Overrides ###############################################################
    def __init__(self, jss, data, **kwargs):
        self.jss = jss
        self._basic_identity = Identity(name="", id="")
        self.kwargs = {}

        if isinstance(data, string_types):
            self._new(data, **kwargs)
            self.cached = "Unsaved"

        elif isinstance(data, ElementTree.Element):
            # Create a new object from passed XML.
            super(Container, self).__init__(jss, data)
            # If this has an ID, assume it's from the JSS and set the
            # cache time, otherwise set it to "Unsaved".
            if data.findtext("id") or data.findtext("general/id"):
                self.cached = dt.datetime.now()
            else:
                self.cached = "Unsaved"

        elif isinstance(data, Identity):
            # This is basic identity information, probably from a
            # listing operation.
            new_xml = PrettyElement(tag=self.root_tag)
            super(Container, self).__init__(jss, new_xml)
            self._basic_identity = Identity(data)
            # Store any kwargs used in retrieving this object.
            self.kwargs = kwargs

        else:
            raise TypeError(
                "JSSObjects data argument must be of type "
                "xml.etree.ElemenTree.Element, Identity, or str")

    def __repr__(self):
        return "<{} with id: {} name: {} cached: {} at 0x{:0x}>".format(
            self.__class__.__name__, self.id, self.name, self.cached,
            id(self))

    def __contains__(self, obj):
        if hasattr(obj, "as_list_data"):
            list_data = obj.as_list_data()
        else:
            return False
        other_id = list_data.findtext("id")
        tags = self.iter(list_data.tag)
        # Give findtext a default non-integer value so that it won't
        # ever compare equal if not found.
        return any(i.findtext("id", "Nay") == other_id for i in tags)

    def retrieve(self, clear_kwargs=False):
        """Replace this object's data with JSS data, reset cache-age.

        Args:
            clear_kwargs (bool): If True, clear the stored request
                kwargs prior to retrieving the full record.
        """
        if clear_kwargs:
            self.kwargs = {}
        super(Container, self).retrieve()

    @classmethod
    def build_query(cls, data, **kwargs):
        """Return the path for query based on data type and contents.

        Args:
            data (int, str, unicode, None): Accepts multiple types.

                Int: Generate URL to object with data ID.
                None: Get basic object GET URL (list).
                String/Unicode: Search for <data> with default_search,
                    usually "name". If the wildcard character, '*' is
                    present, and the object can do 'match' searches,
                    a match search will be performed rather than a
                    normal one.
                String/Unicode with "=": Other searches, for example
                    Computers can be searched by uuid with:
                    "udid=E79E84CB-3227-5C69-A32C-6C45C2E77DF5"
                    See the class "search_types" attribute for options.
            kwargs:
                Some classes allow additional filters, subsets, etc,
                in their queries. Check the object's `allowed_kwargs`
                attribute for a complete list of implemented keys.

                Not all classes offer all types of searches, nor are they
                all necessarily offered in a single query. Consult the
                Casper API documentation for usage.

                In general, the key name is applied to the end of the
                URL, followed by the val; e.g. '<url>/subset/general'.

                subset (list of str or str): XML subelement tags to
                    request (e.g.  ['general', 'purchasing']), OR an '&'
                    delimited string (e.g.  'general&purchasing').
                    Defaults to None.
                start_date/end_date (str or datetime): Either dates in
                    the form 'YYYY-MM-DD' or a datetime.datetime object.

        Returns:
            str path construction for this class to query.
        """
        url_components = ['JSSResource', cls._endpoint_path]

        try:
            data = int(data)
        except (ValueError, TypeError):
            pass
        if isinstance(data, int):
            url_components.extend([cls._id_path, str(data)])

        elif isinstance(data, string_types):
            if "=" in data:
                key, value = data.split("=")   # pylint: disable=no-member
                if key in cls.search_types:
                    url_components.extend([cls.search_types[key], value])

                else:
                    raise TypeError(
                        "This object cannot be queried by %s." % key)

            elif "*" in data and _MATCH in cls.search_types:
                # If wildcard char present, make this a match search if
                # possible
                url_components.extend([cls.search_types[_MATCH], data])
            elif data:
                url_components.extend(
                    [cls.search_types[cls.default_search], data])

        url_components.extend(cls._process_kwargs(kwargs))

        url = os.path.join(*url_components)

        return url

    @classmethod
    def _process_kwargs(cls, kwargs):
        kwarg_urls = []
        if kwargs and all(key in cls.allowed_kwargs for key in kwargs):
            kwargs = cls._handle_kwargs(kwargs)
            for key, val in kwargs.items():
                kwarg_urls.extend(cls._urlify_arg(key, val))
        return kwarg_urls

    @classmethod
    def _handle_kwargs(cls, kwargs):
        """Do nothing. Can be overriden by classes which need it."""
        return kwargs

    @classmethod
    def _urlify_arg(cls, key, val):
        """Convert keyword arguments' values to proper format for GET"""
        if key == 'subset':
            if not isinstance(val, list):
                val = val.split("&")

            # If this is not a "basic" subset, and it's missing
            # "general", add it in, because we need the ID.
            if all(k not in val for k in ("general", "basic")):
                val.append("general")

            return ['subset', "&".join(val)]

        elif key == 'date_range':
            start, end = val
            fmt = lambda s: s.strftime('%Y-%m-%d')
            start = start if isinstance(start, string_types) else fmt(end)
            end = end if isinstance(end, string_types) else fmt(end)
            return ['{}_{}'.format(start, end)]

        else:
            return [key, val]

    @property
    def url(self):
        """Return the path subcomponent of the url to this object.

        For example: "computers/id/451"
        """
        url_components = ['JSSResource', self._endpoint_path, self._id_path, self.id]
        url_components.extend(self._process_kwargs(self.kwargs))
        return os.path.join(*url_components)

    def save(self):
        """Update or create a new object on the JSS.

        If this object is not yet on the JSS, this method will create
        a new object with POST, otherwise, it will try to update the
        existing object with PUT.

        Data validation is up to the client; The JSS in most cases will
        at least give you some hints as to what is invalid.
        """
        # Object probably exists if it has an ID (user can't assign
        # one).
        if self.can_put and self.id != "0":
            # The JSS will reject PUT requests for objects that do not have
            # a category. The JSS assigns a name of "No category assigned",
            # which it will reject. Therefore, if that is the category
            # name, changed it to "", which is accepted.
            categories = [elem for elem in self.findall("category")]
            categories.extend([elem for elem in self.findall("category/name")])
            for cat_tag in categories:
                if cat_tag.text == "No category assigned":
                    cat_tag.text = ""

            super(Container, self).save()

        elif self.can_post:
            try:
                id_ = self.jss.post(self.url, data=self)
            except PostError as err:
                raise PostError(err)

            self._basic_identity["id"] = id_
            # Replace current instance's data with new, JSS-validated data.
            # and update cached time.
            self.retrieve()

        else:
            raise MethodNotAllowedError(self.__class__.__name__)

    # Methods #################################################################
    def _new(self, name, **kwargs):
        """Create a new JSSObject with name and "keys".

        Generate a default XML template for this object, based on
        the class attribute "keys".

        Args:
            name: String name of the object to use as the
                object's name property.
            kwargs:
                Accepted keyword args can be viewed by checking the
                "data_keys" class attribute. Typically, they include all
                top-level keys, and non-duplicated keys used elsewhere.

                Values will be cast to string. (Int 10, bool False
                become string values "10" and "false").

                Ignores kwargs that aren't in object's keys attribute.
        """
        new_xml = PrettyElement(tag=self.root_tag)
        super(Container, self).__init__(self.jss, new_xml)
        self.cached = "Unsaved"

        # Name is required, so set it outside of the helper func.
        current_tag = self
        for path_element in self._name_element.split("/"):
            ElementTree.SubElement(current_tag, path_element)
            current_tag = current_tag.find(path_element)

        current_tag.text = name

        for item in self.data_keys.items():
            self._set_xml_from_keys(self, item, **kwargs)

    def _set_xml_from_keys(self, root, item, **kwargs):
        """Create SubElements of root with kwargs.

        Args:
            root: Element to add SubElements to.
            item: Tuple key/value pair from self.data_keys to add.
            kwargs:
                For each item in self.data_keys, if it has a
                corresponding kwarg, create a SubElement at root with
                the kwarg's value.

                Int and bool values will be cast to string. (Int 10,
                bool False become string values "10" and "false").

                Dicts will be recursively added to their key's Element.
        """
        key, val = item
        target_key = root.find(key)
        if target_key is None:
            target_key = ElementTree.SubElement(root, key)

        if isinstance(val, dict):
            for dict_item in val.items():
                self._set_xml_from_keys(target_key, dict_item, **kwargs)
            return

        # Convert kwarg data to the appropriate string.
        # TODO: Factor out repeated kwargs[key] usage
        if key in kwargs:
            kwarg = kwargs[key]
            if isinstance(kwarg, bool):
                kwargs[key] = str(kwargs[key]).lower()
            elif kwarg is None:
                kwargs[key] = ""
            elif isinstance(kwarg, int):
                kwargs[key] = str(kwargs[key])
            elif isinstance(kwarg, JSSObject):
                kwargs[key] = kwargs[key].name

        target_key.text = kwargs.get(key, val)

    @property
    def basic(self):
        """Returns a copy of the basic identity information

        For most objects, this is just ID and name. Computers have a
        'basic' subset that includes some other data as well.

        The returned data is a copy to enforce being read-only.
        """
        return copy.copy(self._basic_identity)

    @property
    def name(self):
        """Return object name or None."""
        if not self.cached:
            # name = self._basic_name
            name = self._basic_identity["name"]
        else:
            name = self.findtext("name") or self.findtext("general/name")
        return name

    @name.setter
    def name(self, name):
        if self.findtext("name"):
            path = "name"
        elif self.findtext("general/name"):
            path = "general/name"
        else:
            raise JSSError("Name property couldn't be found!")
        # self._basic_name = self.find('name').text = name
        self._basic_identity["name"] = self.find('name').text = name

    @property
    def id(self):   # pylint: disable=invalid-name
        """Return object ID or None."""
        # Most objects have ID nested in general. Groups don't.

        # After much consideration, I will treat id's as strings. We
        # can't assign ID's, so there's no need to perform arithmetic on
        # them, and having to convert to str all over the place is
        # gross. str equivalency still works.
        if not self.cached or self.cached == "Unsaved":
            # id_ = self._basic_id
            id_ = self._basic_identity["id"]
        else:
            id_ = self.findtext("id") or self.findtext("general/id")
        # If no ID has been found, this object hasn't been POSTed to the
        # JSS. New objects use the ID "0".
        return id_ or "0"

    def as_list_data(self):
        """Return an Element to be used in a list.

        Most lists want an element with tag of root_tag, and
        subelements of id and name.

        Returns:
            Element: list representation of object.
        """
        element = PrettyElement(self.root_tag)
        id_ = ElementTree.SubElement(element, "id")
        id_.text = self.id
        name = ElementTree.SubElement(element, "name")
        name.text = self.name
        return element

    def delete(self, data=None):
        """Delete this object from the JSS."""
        if not self.can_delete:
            raise MethodNotAllowedError(self.__class__.__name__)
        if data:
            self.jss.delete(self.url, data=data)
        else:
            self.jss.delete(self.url)

    def _handle_location(self, location):
        """Return an element located at location with flexible args.

        Args:
            location: String xpath to use in an Element.find search OR
                an Element (which is simply returned).

        Returns:
            The found Element.

        Raises:
            ValueError if the location is a string that results in a
            find of None.
        """
        if not isinstance(location, ElementTree.Element):
            element = self.find(location)
            if element is None:
                raise ValueError("Invalid path!")
        else:
            element = location
        return element

    def set_bool(self, location, value):
        """Set a boolean value.

        Casper booleans in XML are string literals of "true" or "false".
        This method sets the text value of "location" to the correct
        string representation of a boolean.

        Args:
            location: Element or a string path argument to find()
            value: Boolean or string value to set. (Accepts
            "true"/"True"/"TRUE"; all other strings are False).
        """
        element = self._handle_location(location)
        if isinstance(value, string_types):
            value = True if value.upper() == "TRUE" else False
        elif not isinstance(value, bool):
            raise ValueError
        if value is True:
            element.text = "true"
        else:
            element.text = "false"

    def add_object_to_path(self, obj, location):
        """Add an object of type Container to location.

        This method determines the correct list representation of an
        object and adds it to "location". For example, add a Computer to
        a ComputerGroup. The ComputerGroup will now have a child
        Computers/Computer tag with subelements "name" and "id".

        Args:
            obj: A Container subclass.
            location: Element or a string path argument to find()

        Returns:
            Element for the object just added.
        """
        location = self._handle_location(location)
        location.append(obj.as_list_data())
        results = [item for item in location.getchildren() if
                   item.findtext("id") == obj.id][0]
        return results

    def remove_object_from_list(self, obj, list_element):
        """Remove an object from a list element.

        Args:
            obj: Accepts JSSObjects, id's, and names
            list_element: Accepts an Element or a string path to that
                element
        """
        list_element = self._handle_location(list_element)

        if isinstance(obj, Container):
            results = [item for item in list_element.getchildren() if
                       item.findtext("id") == obj.id]
        elif isinstance(obj, (int, string_types)):
            results = [item for item in list_element.getchildren() if
                       item.findtext("id") == str(obj) or
                       item.findtext("name") == obj]

        if len(results) == 1:
            list_element.remove(results[0])
        elif len(results) > 1:
            raise ValueError("There is more than one matching object at that "
                             "path!")

    def clear_list(self, list_element):
        """Clear an Element or everything below a path.

        For example, to clear all Computers in a ComputerGroup:
            `computer_group_instance.clear_list("computers")`

        Args:
            list_element: Accepts an Element or a string path to that
                element, to remove.
        """
        list_element = self._handle_location(list_element)
        list_element.clear()

    @classmethod
    def from_file(cls, jss, filename):
        """Create a new JSSObject from an external XML file.

        Args:
            jss: A JSS object.
            filename: String path to an XML file.
        """
        tree = ElementTree.parse(filename)
        root = tree.getroot()
        return cls(jss, root)

    @classmethod
    def from_string(cls, jss, xml_string):
        """Creates a new JSSObject from a string or unicode.

        Args:
            jss: A JSS object.
            xml_string (str or unicode): XML file data used to create
                object.
        """
        # ElementTree.fromstring in python2 really wants bytes.
        if isinstance(xml_string, unicode):
            xml_string = xml_string.encode('UTF-8')
        root = ElementTree.fromstring(xml_string)
        return cls(jss, root)

    @classmethod
    def from_pickle(cls, path):
        """Load from pickle file.

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to load from.
                Path will have ~ expanded prior to opening.
        """
        gz_magic = "\x1f\x8b\x08"

        # Determine if file is gzipped.
        with open(os.path.expanduser(path), "rb") as pickle:
            pickle_magic = pickle.read(len(gz_magic))
            compressed = True if pickle_magic == gz_magic else False

        opener = gzip.open if compressed else open

        with opener(os.path.expanduser(path), "rb") as pickle:
            return cPickle.Unpickler(pickle).load()

    def to_pickle(self, path, compress=True):
        """Write this object to a Python Pickle.

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening, and `.gz`
                appended if compress=True and it's missing.
            compress (bool): Whether to gzip output. Default is True.
        """
        path = os.path.expanduser(path)
        gz_ext = ".gz"
        if compress and not path.endswith(gz_ext):
            path = path + gz_ext

        opener = gzip.open if compress else open
        with opener(path, 'wb') as file_handle:
            cPickle.Pickler(
                file_handle, cPickle.HIGHEST_PROTOCOL).dump(self)


class Group(Container):
    """Abstract class for ComputerGroup and MobileDeviceGroup."""

    def add_criterion(self, name, priority, and_or, search_type, value):   # pylint: disable=too-many-arguments
        """Add a search criteria object to a smart group.

        Args:
            name: String Criteria type name (e.g. "Application Title")
            priority: Int or Str number priority of criterion.
            and_or: Str, either "and" or "or".
            search_type: String Criteria search type. (e.g. "is", "is
                not", "member of", etc). Construct a SmartGroup with the
                criteria of interest in the web interface to determine
                what range of values are available.
            value: String value to search for/against.
        """
        criterion = SearchCriteria(name, priority, and_or, search_type, value)
        self.criteria.append(criterion)

    @property
    def is_smart(self):
        """Returns boolean for whether group is Smart."""
        result = False
        if self.findtext("is_smart") == "true":
            result = True
        return result

    @is_smart.setter
    def is_smart(self, value):
        """Set group is_smart property to value.

        Args:
            value: Boolean.
        """
        self.set_bool("is_smart", value)
        if value is True:
            if self.find("criteria") is None:
                # pylint: disable=attribute-defined-outside-init
                self.criteria = ElementTree.SubElement(self, "criteria")
                # pylint: enable=attribute-defined-outside-init

    def add_device(self, device, container):
        """Add a device to a group. Wraps JSSObject.add_object_to_path.

        Args:
            device: A JSSObject to add (as list data), to this object.
            location: Element or a string path argument to find()
        """
        # There is a size tag which the JSS manages for us, so we can
        # ignore it.
        if self.findtext("is_smart") == "false":
            self.add_object_to_path(device, container)
        else:
            # Technically this isn't true. It will strangely accept
            # them, and they even show up as members of the group!
            raise ValueError("Devices may not be added to smart groups.")

    def has_member(self, device_object):
        """Return bool whether group has a device as a member.

        Args:
            device_object (Computer or MobileDevice). Membership is
            determined by ID, as names can be shared amongst devices.
        """
        if device_object.tag == "computer":
            container_search = "computers/computer"
        elif device_object.tag == "mobile_device":
            container_search = "mobile_devices/mobile_device"
        else:
            raise ValueError

        return len([device for device in self.findall(container_search) if
                    device.findtext("id") == device_object.id]) is not 0


# class Scoped(Container):
#     """Abstract class for a container that supports a <scope> element."""
#
#     def add_object_to_scope(self, obj):
#         """Add an object to the appropriate scope block.
#
#         Args:
#             obj: JSSObject to add to scope. Accepted subclasses are:
#                 Computer
#                 ComputerGroup
#                 Building
#                 Department
#
#         Raises:
#             TypeError if invalid obj type is provided.
#         """
#         if isinstance(obj, Computer):
#             self.add_object_to_path(obj, "scope/computers")
#         elif isinstance(obj, ComputerGroup):
#             self.add_object_to_path(obj, "scope/computer_groups")
#         elif isinstance(obj, Building):
#             self.add_object_to_path(obj, "scope/buildings")
#         elif isinstance(obj, Department):
#             self.add_object_to_path(obj, "scope/departments")
#         else:
#             raise TypeError
#
#     def clear_scope(self):
#         """Clear all objects from the scope, including exclusions."""
#         clear_list = ["computers", "computer_groups", "buildings",
#                       "departments", "limit_to_users/user_groups",
#                       "limitations/users", "limitations/user_groups",
#                       "limitations/network_segments", "exclusions/computers",
#                       "exclusions/computer_groups", "exclusions/buildings",
#                       "exclusions/departments", "exclusions/users",
#                       "exclusions/user_groups", "exclusions/network_segments"]
#         for section in clear_list:
#             self.clear_list("%s%s" % ("scope/", section))
#
#     def add_object_to_exclusions(self, obj):
#         """Add an object to the appropriate scope exclusions
#         block.
#
#         Args:
#             obj: JSSObject to add to exclusions. Accepted subclasses
#                     are:
#                 Computer
#                 ComputerGroup
#                 Building
#                 Department
#
#         Raises:
#             TypeError if invalid obj type is provided.
#         """
#         if isinstance(obj, Computer):
#             self.add_object_to_path(obj, "scope/exclusions/computers")
#         elif isinstance(obj, ComputerGroup):
#             self.add_object_to_path(obj, "scope/exclusions/computer_groups")
#         elif isinstance(obj, Building):
#             self.add_object_to_path(obj, "scope/exclusions/buildings")
#         elif isinstance(obj, Department):
#             self.add_object_to_path(obj, "scope/exclusions/departments")
#         else:
#             raise TypeError
#
#     def add_object_to_limitations(self, obj):
#         """Add an object to the appropriate scope limitations
#         block.
#
#         Args:
#             obj: JSSObject to add to limitations. Accepted subclasses
#                 are:
#                     User
#                     UserGroup
#                     NetworkSegment
#                     IBeacon
#
#         Raises:
#             TypeError if invalid obj type is provided.
#         """
#         if isinstance(obj, User):
#             self.add_object_to_path(obj, "scope/limitations/users")
#         elif isinstance(obj, UserGroup):
#             self.add_object_to_path(obj, "scope/limitations/user_groups")
#         elif isinstance(obj, NetworkSegment):
#             self.add_object_to_path(obj, "scope/limitations/network_segments")
#         elif isinstance(obj, IBeacon):
#             self.add_object_to_path(obj, "scope/limitations/ibeacons")
#         else:
#             raise TypeError
#

class SearchCriteria(PrettyElement):
    """Object for encapsulating a smart group search criteria."""
    root_tag = "criterion"

    def __init__(self, name, priority, and_or, search_type, value):   # pylint: disable=too-many-arguments
        """Init a SearchCriteria.

        Args:
            name: String Criteria type name (e.g. "Application Title")
            priority: Int or Str number priority of criterion.
            and_or: Str, either "and" or "or".
            search_type: String Criteria search type. (e.g. "is", "is
                not", "member of", etc). Construct a SmartGroup with the
                criteria of interest in the web interface to determine
                what range of values are available.
            value: String value to search for/against.
        """
        super(SearchCriteria, self).__init__(tag=self.root_tag)
        crit_name = ElementTree.SubElement(self, "name")
        crit_name.text = name
        crit_priority = ElementTree.SubElement(self, "priority")
        crit_priority.text = str(priority)
        crit_and_or = ElementTree.SubElement(self, "and_or")
        crit_and_or.text = and_or
        crit_search_type = ElementTree.SubElement(self, "search_type")
        crit_search_type.text = search_type
        crit_value = ElementTree.SubElement(self, "value")
        crit_value.text = value
